'''
Created on Oct 10, 2012

@author: smirarab
'''
from sepp.algorithm import AbstractAlgorithm
from sepp.config import options
from sepp.tree import PhylogeneticTree
from sepp.alignment import MutableAlignment, ExtendedAlignment,\
    hamming_distance
from sepp.problem import SeppProblem
from dendropy.datamodel.treemodel import Tree
from sepp.jobs import HMMBuildJob, HMMSearchJob, HMMAlignJob, PplacerJob,\
    MergeJsonJob
from sepp.scheduler import JobPool, Join
from sepp import get_logger
from sepp.math_utils import lcm
_LOG = get_logger(__name__)


def get_placement_job_name(chunk_number):
    return "pplacer_%d" % chunk_number


class JoinSearchJobs(Join):
    '''
    After all search jobs have finished on tips, we need to figure out which
    fragment goes  to which subset and start aligning fragments.
    This join takes care of that step.
    '''
    def __init__(self):
        Join.__init__(self)

    def setup_with_root_problem(self, root_problem):
        self.root_problem = root_problem
        for p in root_problem.iter_leaves():
            self.add_job(p.jobs["hmmsearch"])

    def figureout_fragment_subset(self):
        ''' Figure out which fragment should go to which subproblem'''
        if "fragments.distribution.done" in self.root_problem.annotations:
            return
        max_evalues = dict(
            [(name, (None, None)) for name in
             self.root_problem.fragments.keys()])
        for fragment_chunk_problem in self.root_problem.iter_leaves():
            align_problem = fragment_chunk_problem.get_parent()
            assert isinstance(align_problem, SeppProblem)
            '''For each subproblem start with an empty set of fragments,
            and add to them as we encounter new best hits for that
            subproblem'''
            if align_problem.fragments is None:
                align_problem.fragments = \
                    self.root_problem.fragments.get_soft_sub_alignment([])
            search_res = fragment_chunk_problem.get_job_result_by_name(
                "hmmsearch")
            for key, val in search_res.items():
                (best_value, prev_align_problem) = max_evalues[key]
                ''' If this is better than previous best hit, remove this
                fragment from the previous hit, and add it to this subproblem
                '''
                if best_value is None or (best_value < val[1]):
                    max_evalues[key] = (val[1], align_problem)

        # TODO: is the following efficient enough? Do we need to make lists
        # and then turn them to sets?
        notScored = []
        for key, v in max_evalues.items():
            if v[1] is None:
                notScored.append(key)
            else:
                v[1].fragments.seq_names.add(key)

        self.root_problem.annotations["fragments.distribution.done"] = 1

        ''' Make sure all fragments are in at least one subproblem.
        TODO: what to do with those that are not?  For now, only output
        warning message'''
        # notScored = [k for k, v in max_evalues.iteritems() if v[1] is None]
        _LOG.warning(
            "Fragments %s are not scored against any subset" % str(notScored))
        # assert len(notScored) == 0, "Fragments %s are not scored against
        # any subset" %str(notScored)

    def perform(self):
        '''
        Distributes fragments to alignments subsets with best score,
        and runs align jobs on those. Also, creates new chunks of fragments
        for better parallelism.
        '''

        ''' Figure out which fragment should go to which subproblem'''
        self.figureout_fragment_subset()

        ''' For each alignment subproblem,
        1) make sure its fragments are evenly distributed to fragment chunks.
        2) Setup alignment jobs for its children and enqueue them'''
        alg_problems = [alg for p in self.root_problem.children
                        for alg in p.children]
        for alg_problem in alg_problems:
            assert isinstance(alg_problem, SeppProblem)
            chunks = len(alg_problem.get_children())
            fragment_chunks = alg_problem.fragments.divide_to_equal_chunks(
                chunks)

            ''' Now setup alignment jobs and enqueue them'''
            for (i, fragment_chunk_problem) in enumerate(alg_problem.children):
                fragment_chunk_problem.fragments = fragment_chunks[i]
                aj = fragment_chunk_problem.jobs['hmmalign']
                assert isinstance(aj, HMMAlignJob)
                ''' First Complete setting up alignments'''
                aj.hmmmodel = alg_problem.get_job_result_by_name('hmmbuild')
                aj.base_alignment = alg_problem.jobs["hmmbuild"].infile

                if (
                    fragment_chunk_problem.fragments is None or
                    fragment_chunk_problem.fragments.is_empty()
                   ):
                    aj.fake_run = True
                else:
                    fragment_chunk_problem.fragments.write_to_path(
                        aj.fragments)
                ''' Now the align job can be put on the queue '''
                JobPool().enqueue_job(aj)

    def __str__(self):
        return "join search jobs for all tips of ", self.root_problem


class JoinAlignJobs(Join):
    '''
    After all alignments jobs for a placement subset have finished,
    we need to build those extended alignments and start placing fragments.
    This join takes care of that step.
    '''
    def __init__(self):
        Join.__init__(self)

    def setup_with_placement_problem(self, placement_problem):
        self.placement_problem = placement_problem
        self.root_problem = placement_problem.parent
        for p in placement_problem.iter_leaves():
            self.add_job(p.jobs["hmmalign"])

    def merge_subalignments(self):
        '''
        Merge alignment subset extended alignments to get one extended
        alignment for current placement subset.
        '''
        pp = self.placement_problem
        _LOG.info("Merging sub-alignments for placement problem : %s." %
                  (pp.label))
        ''' First find fragments assigned to this placement problem'''
        pp.fragments = pp.parent.fragments.get_soft_sub_alignment([])
        for ap in pp.get_children():
            pp.fragments.seq_names |= set(ap.fragments)

        ''' Then, gather a list of all alignments relevant to this placement
        subset'''
        fragfilesperap = dict()
        for ap in pp.children:
            assert isinstance(ap, SeppProblem)
            ''' Get all fragment chunk alignments for this alignment subset'''
            aligned_files = [fp.get_job_result_by_name('hmmalign') for
                             fp in ap.children]
            fragfilesperap[ap] = aligned_files

        ''' Now, build an extended alignment *per each fragment chunk*.
            Simply merge all hmmalign results for fragment chunk numbered i'''
        extendedAlignments = []
        for i in range(0, self.root_problem.fragment_chunks):
            extendedAlignment = ExtendedAlignment(pp.fragments.seq_names)
            for ap in pp.children:
                # _LOG.debug("Merging fragment chunks for subalignment : %s."
                # %(ap.label))
                if fragfilesperap[ap][i]:
                    ap_alg = ap.read_extendend_alignment_and_relabel_columns(
                        ap.jobs["hmmbuild"].infile, [fragfilesperap[ap][i]])
                else:
                    ap_alg = ap.read_extendend_alignment_and_relabel_columns(
                        ap.jobs["hmmbuild"].infile, [])
                _LOG.debug(
                    ("Merging alignment subset into placement subset for "
                     "chunk %d: %s.") % (i, ap.label))
                extendedAlignment.merge_in(ap_alg, convert_to_string=False)
            '''Extended alignmnts have all fragments. remove the ones that
               don't belong to thsi chunk'''
            extendedAlignment.remove_missing_fragments()
            extendedAlignment.from_bytearray_to_string()
            extendedAlignments.append(extendedAlignment)
        return extendedAlignments

    def perform(self):
        pp = self.placement_problem
        fullExtendedAlignments = self.merge_subalignments()

        for i in range(0, self.root_problem.fragment_chunks):
            fullExtendedAlignment = fullExtendedAlignments[i]
            # Split the backbone alignment and query sequences
            # into separate files
            queryExtendedAlignment = \
                fullExtendedAlignment.get_fragments_readonly_alignment()
            baseAlignment = fullExtendedAlignment.get_base_readonly_alignment()
            pj = pp.jobs[get_placement_job_name(i)]
            assert isinstance(pj, PplacerJob)
            if (queryExtendedAlignment.is_empty()):
                pj.fake_run = True

            # Write out the extended alignments, split into query and full-
            # length for pplacer
            queryExtendedAlignment.write_to_path(pj.extended_alignment_file)
            baseAlignment.write_to_path(pj.backbone_alignment_file)

            # But keep the extended alignment on everything
            pj.set_attribute(
                "full_extended_alignment_object", fullExtendedAlignment)

            JobPool().enqueue_job(pj)

    def __str__(self):
        return "join align jobs for tips of ", self.placement_problem


class ExhaustiveAlgorithm(AbstractAlgorithm):
    '''
    This implements the exhaustive algorithm where all alignments subsets
    are searched for every fragment.
    '''
    def __init__(self):
        AbstractAlgorithm.__init__(self)
        self.place_nomatch_fragments = False
        ''' Hardcoded E-Lim for hmmsearch '''  # TODO: what to do with this
        self.elim = 99999999
        self.filters = False
        self.symfrac = True
        self.strategy = options().exhaustive.strategy
        self.decomp_strategy = options().decomp_strategy
        self.minsubsetsize = int(options().exhaustive.minsubsetsize)
        # Temp fix for now,
        self.molecule = self.options.molecule
        self.distances = dict()

    def compute_distances(self, sequences):
        for seq1, val1 in sequences.items():
            for seq2, val2 in sequences.items():
                if ("".join([seq1, seq2]) not in self.distances):
                    self.distances["".join([seq1, seq2])] = \
                        hamming_distance(val1, val2)
                    self.distances["".join([seq2, seq1])] = \
                        self.distances["".join([seq1, seq2])]

    def merge_results(self):
        assert isinstance(self.root_problem, SeppProblem)

        '''Generate single extended alignment'''
        fullExtendedAlignment = ExtendedAlignment(
            self.root_problem.fragments.keys())
        # self.root_problem.get_children()[0].jobs[get_placement_job_name(0)]\
        # .get_attribute("full_extended_alignment_object")
        for pp in self.root_problem.get_children():
            for i in range(0, self.root_problem.fragment_chunks):
                extended_alignment = pp.jobs[
                    get_placement_job_name(i)].get_attribute(
                        "full_extended_alignment_object")
                fullExtendedAlignment.merge_in(
                    extended_alignment, convert_to_string=True)
        self.results = fullExtendedAlignment

        mergeinput = []
        '''Append main tree to merge input'''
        mergeinput.append(
            "%s;" % (self.root_problem.subtree.compose_newick(labels=True)))
        for pp in self.root_problem.get_children():
            assert isinstance(pp, SeppProblem)
            for i in range(0, self.root_problem.fragment_chunks):
                if (pp.get_job_result_by_name(
                        get_placement_job_name(i)) is None):
                    continue
                '''Append subset trees and json locations to merge input'''
                mergeinput.append(
                    "%s;\n%s" % (
                        pp.subtree.compose_newick(labels=True),
                        pp.get_job_result_by_name(get_placement_job_name(i))))
        mergeinput.append("")
        mergeinput.append("")
        meregeinputstring = "\n".join(mergeinput)
        mergeJsonJob = MergeJsonJob()
        mergeJsonJob.setup(meregeinputstring,
                           self.get_output_filename("placement.json"))
        mergeJsonJob.run()

    def output_results(self):
        ''' Merged json file is already saved in merge_results function and
            full extended alignment already created in merge_results function
        '''
        outfilename = self.get_output_filename("alignment.fasta")
        self.results .write_to_path(outfilename)
        self.results.remove_insertion_columns()
        outfilename = self.get_output_filename("alignment_masked.fasta")
        self.results.write_to_path(outfilename)
        namerev_script = self.root_problem.subtree.rename_script()
        if namerev_script:
            outfilename = self.get_output_filename("rename-json.py")
            with open(outfilename, 'w') as s:
                s.write(namerev_script)

    def check_options(self, supply=[]):
        if (options().info_file is None):
            supply = supply + ["raxml file"]
        AbstractAlgorithm.check_options(self, supply)

    def modify_tree(self, a_tree):
        pass

    def build_subproblems(self):
        (alignment, tree) = self.read_alignment_and_tree()

        if options().distance != 1:
            self.compute_distances(alignment)

        assert isinstance(tree, PhylogeneticTree)
        assert isinstance(alignment, MutableAlignment)

        tree.get_tree().resolve_polytomies()
        # Label edges with numbers so that we could assemble things back
        # at the end
        tree.lable_edges()

        ''' Make sure size values are set, and are meaningful. '''
        self.check_and_set_sizes(alignment.get_num_taxa())

        self._create_root_problem(tree, alignment)

        ''' Decompose the tree based on placement subsets'''
        placement_tree_map = PhylogeneticTree(
            Tree(tree.den_tree)).decompose_tree(
                self.options.placement_size,
                strategy=self.strategy,
                minSize=self.options.placement_size / int(
                    self.options.exhaustive.placementminsubsetsizefacotr),
                tree_map={}, pdistance=1,
                decomp_strategy=self.decomp_strategy,
                distances=self.distances,
                maxDiam=None)
        assert len(placement_tree_map) > 0, (
            "Tree could not be decomposed"
            " given the following settings; strategy:%s minsubsetsize:%s"
            " placement_size:%s"
            % (self.strategy, self.minsubsetsize, self.options.placement_size))
        _LOG.info("Breaking into %d placement subsets." % len(
            placement_tree_map))

        ''' For placement subsets create a placement subproblem,
            and decompose further'''
        for (p_key, p_tree) in placement_tree_map.items():
            assert isinstance(p_tree, PhylogeneticTree)
            placement_problem = SeppProblem(
                p_tree.leaf_node_names(), self.root_problem)
            placement_problem.subtree = p_tree
            placement_problem.label = "P_%s" % str(p_key)
            _LOG.debug(
                "Placement subset %s has %d nodes" %
                (placement_problem.label, len(p_tree.leaf_node_names())))
            ''' Further decompose to alignment subsets '''
            alignment_tree_map = PhylogeneticTree(
                Tree(p_tree.den_tree)).decompose_tree(
                    self.options.alignment_size,
                    strategy=self.strategy,
                    minSize=self.minsubsetsize,
                    tree_map={},
                    decomp_strategy=self.options.decomp_strategy,
                    pdistance=options().distance,
                    distances=self.distances,
                    maxDiam=self.options.maxDiam)
            assert len(alignment_tree_map) > 0, (
                "Tree could not be decomposed"
                " given the following settings; strategy:%s"
                " minsubsetsize:%s alignmet_size:%s" %
                (self.strategy, self.minsubsetsize,
                 self.options.alignment_size))

            _LOG.debug("Placement subset %s has %d alignment subsets: %s" %
                       (placement_problem.label, len(alignment_tree_map),
                        str(sorted(alignment_tree_map.keys()))))
            _LOG.debug("Placement subset %s has %d taxa:" %
                       (placement_problem.label,
                        sum([len(a_tree.leaf_node_names())
                             for a_tree in alignment_tree_map.values()])))
            for (a_key, a_tree) in alignment_tree_map.items():
                assert isinstance(a_tree, PhylogeneticTree)
                self.modify_tree(a_tree)
                alignment_problem = SeppProblem(a_tree.leaf_node_names(),
                                                placement_problem)
                alignment_problem.subtree = a_tree
                alignment_problem.label = "A_%s_%s" % (str(p_key), str(a_key))

        _LOG.info("Breaking into %d alignment subsets." %
                  (len(list(self.root_problem.iter_leaves()))))

        ''' Divide fragments into chunks, to help achieve better parallelism'''
        fragment_chunk_files = self.create_fragment_files()
        self.root_problem.fragment_chunks = len(fragment_chunk_files)
        for alignment_problem in self.root_problem.iter_leaves():
            for afc in range(0, self.root_problem.fragment_chunks):
                frag_chunk_problem = SeppProblem(alignment_problem.taxa,
                                                 alignment_problem)
                frag_chunk_problem.subtree = alignment_problem.subtree
                frag_chunk_problem.label = alignment_problem.label.replace(
                    "A_", "FC_") + "_" + str(afc)
                frag_chunk_problem.fragments = fragment_chunk_files[afc]

        _LOG.info("Breaking each alignment subset into %d fragment chunks." %
                  self.root_problem.fragment_chunks)
        _LOG.debug("Subproblem structure: %s" % str(self.root_problem))
        return self.root_problem

    def create_fragment_files(self):
        alg_subset_count = len(list(self.root_problem.iter_leaves()))
        frag_chunk_count = lcm(
            alg_subset_count, self.options.cpu) // alg_subset_count
        return self.read_and_divide_fragments(frag_chunk_count)

    def _get_new_Join_Align_Job(self):
        return JoinAlignJobs()

    def build_jobs(self):
        assert isinstance(self.root_problem, SeppProblem)
        for placement_problem in self.root_problem.get_children():
            ''' Create placer jobs'''
            for i in range(0, self.root_problem.fragment_chunks):
                pj = PplacerJob()
                pj.partial_setup_for_subproblem(
                    placement_problem, self.options.info_file, i)
                placement_problem.add_job(get_placement_job_name(i), pj)

            '''For each alignment subproblem, ...'''
            for alg_problem in placement_problem.children:
                assert isinstance(alg_problem, SeppProblem)
                ''' create the build model job'''
                bj = HMMBuildJob()
                bj.setup_for_subproblem(
                    alg_problem, symfrac=self.symfrac,
                    molecule=self.molecule,
                    **vars(self.options.hmmbuild))
                alg_problem.add_job(bj.job_type, bj)
                ''' create the search jobs'''
                for fc_problem in alg_problem.get_children():
                    sj = HMMSearchJob()
                    sj.partial_setup_for_subproblem(
                        fc_problem.fragments,
                        fc_problem, self.elim, self.filters)
                    fc_problem.add_job(sj.job_type, sj)
                    ''' create the align job'''
                    aj = HMMAlignJob()
                    fc_problem.add_job(aj.job_type, aj)
                    aj.partial_setup_for_subproblem(
                        fc_problem, molecule=self.molecule)

    def connect_jobs(self):
        ''' a callback function called after hmmbuild jobs are finished'''
        def enq_job_searchfragment(result, search_job):
            search_job.hmmmodel = result
            JobPool().enqueue_job(search_job)
        assert isinstance(self.root_problem, SeppProblem)
        for placement_problem in self.root_problem.get_children():
            '''For each alignment subproblem, ...'''
            for alg_problem in placement_problem.children:
                assert isinstance(alg_problem, SeppProblem)
                ''' create the build model job'''
                bj = alg_problem.jobs["hmmbuild"]
                ''' create the search jobs'''
                for fc_problem in alg_problem.get_children():
                    sj = fc_problem.jobs["hmmsearch"]
                    ''' connect bulid and search jobs'''
                    bj.add_call_Back(
                        lambda result, next_job=sj: enq_job_searchfragment(
                            result, next_job))
            '''Join all align jobs of a placement subset (enqueues
                placement job)'''
            jaj = self._get_new_Join_Align_Job()
            jaj.setup_with_placement_problem(placement_problem)
        ''' Join all search jobs together (enqueues align jobs)'''
        jsj = JoinSearchJobs()
        jsj.setup_with_root_problem(self.root_problem)

    def enqueue_firstlevel_job(self):
        for p in self.root_problem.children:
            for ap in p.children:
                JobPool().enqueue_job(ap.jobs["hmmbuild"])


if __name__ == '__main__':
    ExhaustiveAlgorithm().run()



'''
Created on Oct 2, 2012

@author: smirarab
'''
import unittest
from sepp.jobs import ExternalSeppJob
from sepp.scheduler import JobPool, JobError
from multiprocessing import cpu_count


class TestExternalJob(ExternalSeppJob):
    def tearDown(self):
        # clean up JobPool for other unit tests
        JobPool().terminate()
        JobPool().__init__(cpu_count())

    def __init__(self, pipe=0, **kwargs):
        self.pipe = pipe
        if pipe == 0:
            '''By default errors and output is piped to this process.
            output/err will be available at as self.stdoutdata and
            self.stderrdata and also will be logged'''
            pass
        elif pipe == 1:
            '''The following line forces output to be saved to a file. Note
            that
            1- A file object/descriptor cannot be used here because this is
               executed in parent process, and in child process those file
               handlers will not be available.
            2- self.stdoutdata and self.stderrdata will not be automatically
               filled (to avoid unnecessary I/O). Read stderr and stdout in
               read_results if necessary.
            '''
            kwargs['stdout'] = "./.stdout"
            kwargs['stderr'] = "./.stderr"

        '''always call parent constructor'''
        ExternalSeppJob.__init__(self, "test_find_job", path='dummy', **kwargs)

        '''inputs that need to be set before running this job'''
        self.pattern = None
        self.options = None

    def get_invocation(self):
        return ['find', self.pattern] + (
            [] if self.options is None else self.options.split())

    def characterize_input(self):
        return "pattern: %s options: %s (%s)" % (
            self.pattern, self.options, "pipe: %d" % self.pipe)

    def read_results(self):
        if self.pipe == 2:
            return open("./.stdout").read()
        else:
            if self.stdoutdata is not None:
                return self.stdoutdata
            else:
                return "Nothing captured"


class Test(unittest.TestCase):
    def tearDown(self):
        # clean up JobPool for other unit tests
        JobPool().terminate()
        JobPool().__init__(cpu_count())

    def testSuccess(self):
        find_job = TestExternalJob()
        find_job.pattern = "."
        find_job.options = "-name test*.py"
        JobPool().enqueue_job(find_job)

        JobPool().wait_for_all_jobs()

        res0 = find_job.result.split('\n')[0]

        assert res0 != ""

        find_job = TestExternalJob()
        find_job.pattern = ".."
        find_job.options = "-name %s" % res0.split('/')[-1]
        JobPool().enqueue_job(find_job)

        JobPool().wait_for_all_jobs()

        res1 = find_job.result.split('\n')[0]

        assert res1 != ""

        assert res1.endswith(res0[2:])

    def testError(self):
        find_job = TestExternalJob()
        find_job.pattern = "somerandomdirectorywewillneverhavehere_ordowe"
        ''' Let's ignore this error in subsequent test cases. '''
        find_job.ignore_error = True
        try:
            JobPool().enqueue_job(find_job)
            assert JobPool().wait_for_all_jobs() is False
        except JobError as e:
            assert str(e).find("No such file or directory") != -1, \
                "The error we expected is no such file or directory"

        assert JobPool().get_asynch_result_object(find_job).successful() is \
            False, "We expected the job to fail"

    def testNoPipe(self):
        find_job = TestExternalJob(pipe=1)
        find_job.pattern = "."
        find_job.options = "-name *.py"
        JobPool().enqueue_job(find_job)
        JobPool().wait_for_all_jobs()
        res0 = find_job.result.split('\n')[0]
        assert res0 != ""

        find_job = TestExternalJob(pipe=1)
        find_job.pattern = "somerandomdirectorywewillneverhavehere_ordowe"
        ''' Let's ignore this error in subsequent test cases. '''
        find_job.ignore_error = True
        try:
            JobPool().enqueue_job(find_job)
            assert JobPool().wait_for_all_jobs() is False
            JobPool().get_asynch_result_object(find_job).get()
        except JobError as e:
            assert str(e).find("No such file or directory") != -1, \
                "The error we expected is no such file or directory"

        assert JobPool().get_asynch_result_object(find_job).successful() is \
            False, "We expected the job to fail"


if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testBaseExternalJob']
    unittest.main()



'''
Created on Nov 26, 2012

@author: smirarab
'''
import os
import pickle
import sys
import threading
from sepp import get_logger
import stat
from sepp.filemgr import get_temp_file, get_root_temp_dir, set_root_temp_dir
import gzip
import time
from sepp.config import options
import datetime
# from threading import Lock


_LOG = get_logger(__name__)


class CheckPointState(object):
    '''
    The current state as relevant to the checkpointign feature
    '''

    def __init__(self):
        '''
        Constructor
        '''
        self.root_problem = None
        self.temp_root = None
        self.cumulative_time = 0


def save_checkpoint(checkpoint_manager):
    '''
    This is the callback function that is called periodically to save the
    current state of the system.
    '''
    # Note: this module is not bullet proof in terms of race conditions.
    # Most importantly, it is possible (though extremely unlikely) that
    # while the new temp path is being written (f.write...)
    if checkpoint_manager.is_checkpointing:
        # checkpoint_manager.lock.acquire()
        checkpoint_manager.saving = True
        newTmpDest = get_temp_file("dump", "checkpoints")
        _LOG.info("Checkpoint is being updated: %s" % newTmpDest)
        oldTmpFile = open(checkpoint_manager.checkpoint_path).readlines()
        oldTmpFile = None if len(oldTmpFile) == 0 else oldTmpFile[-1].split(
            ",")[0]

        checkpoint_manager.update_time()

        currenlimit = sys.getrecursionlimit()
        sys.setrecursionlimit(100000)
        picklefile = gzip.GzipFile(newTmpDest, 'wb')
        pickle.dump(checkpoint_manager.checkpoint_state, picklefile, 2)
        picklefile.close()
        sys.setrecursionlimit(currenlimit)

        f = open(checkpoint_manager.checkpoint_path, "a")
        f.write("%s, %s\n" % (newTmpDest, datetime.datetime.now()))
        f.close()
        if oldTmpFile is not None:
            os.remove(oldTmpFile)
        _LOG.info("Checkpoint Saved to: %s and linked in %s." % (
            newTmpDest, checkpoint_manager.checkpoint_path))
        checkpoint_manager.saving = False
        # checkpoint_manager.lock.release()
        checkpoint_manager.timer = threading.Timer(
            options().checkpoint_interval, save_checkpoint,
            args=[checkpoint_manager])
        checkpoint_manager.timer.setDaemon(True)
        checkpoint_manager.timer.start()


class CheckPointManager:
    def __init__(self, checkpoint_file):
        self.checkpoint_path = checkpoint_file
        self.checkpoint_state = None
        self.is_recovering = False
        self.is_checkpointing = False
        self.timer = None
        self._init_state_and_file()
        self.last_checkpoint_time = time.time()
        self.saving = False
        # self.lock = Lock()

    def _init_state_and_file(self):
        if self.checkpoint_path is None:
            return
        if ((not os.path.exists(self.checkpoint_path)) or
           (os.stat(self.checkpoint_path)[stat.ST_SIZE] == 0)):
            open(self.checkpoint_path, "w").close()
            self.checkpoint_state = CheckPointState()
            self.is_checkpointing = True
        else:
            self.is_recovering = True
            self.is_checkpointing = True

    def restore_checkpoint(self):
        _LOG.info("Checkpoint is being restored: %s" %
                  str(self.checkpoint_path))
        assert os.path.exists(self.checkpoint_path)
        lastPath = open(self.checkpoint_path).readlines()[-1].split(",")[0]
        picklefile = gzip.GzipFile(lastPath, 'rb')
        self.checkpoint_state = pickle.load(picklefile)
        picklefile.close()
        set_root_temp_dir(self.checkpoint_state.temp_root)
        _LOG.info("Checkpoint restore finished: %s" % lastPath)

    def remove_checkpoint_file(self):
        if (self.checkpoint_path is not None and
           os.path.exists(self.checkpoint_path)):
            os.remove(self.checkpoint_path)

    def start_checkpointing(self, root_problem):
        if self.is_checkpointing:
            _LOG.info("Checkpoint every %d seconds" %
                      options().checkpoint_interval)
            self.checkpoint_state.root_problem = root_problem
            self.checkpoint_state.temp_root = get_root_temp_dir()
            if self.checkpoint_state.cumulative_time is None:
                self.checkpoint_state.cumulative_time = 0
            save_checkpoint(self)

    def stop_checkpointing(self):
        if self.is_checkpointing:
            self.is_checkpointing = False
            if self.timer is not None:
                self.timer.cancel()
            self.remove_checkpoint_file()
            self.update_time()
            _LOG.info("Stop Checkpointing. Cumulative time: %d" %
                      self.checkpoint_state.cumulative_time)

    def pause_checkpointing(self):
        if self.is_checkpointing:
            if self.timer is not None:
                if self.saving:
                    self.timer.join()
                    self.timer.cancel()
                else:
                    self.timer.cancel()
            _LOG.info("Paused Checkpointing. Cumulative time: %d" %
                      self.checkpoint_state.cumulative_time)

    def update_time(self):
        if self.is_checkpointing:
            self.checkpoint_state.cumulative_time += (
                time.time() - self.last_checkpoint_time)
        self.last_checkpoint_time = time.time()

    def get_total_time(self):
        if self.checkpoint_state is not None:
            return self.checkpoint_state.cumulative_time + time.time() -\
                self.last_checkpoint_time
        else:
            return time.time() - self.last_checkpoint_time

#    def force_checkpoint(self):
#        if self.is_checkpointing:
#            self.timer = threading.Timer(0, save_checkpoint,
#                                         args=[self,False])
#            self.timer.setDaemon(True)
#            self.timer.start()
#            self.timer.join()



'''
Created on May 13, 2014
@author: namphuon
'''
import argparse
import os
from sepp.alignment import MutableAlignment


def parse_args():
    parser = argparse.ArgumentParser(
        description='Performs various tools for TIPP.')
    parser.add_argument(
        '-t', '--threshold', default=None, metavar='THRESHOLD',
        help='Split based on this threshold of length',
        type=int, dest='threshold')
    parser.add_argument(
        '-i', '--input', default=None, metavar='INPUT',
        help='INPUT sequence file (default=None)', required=True,
        type=str, dest='input')
    parser.add_argument(
        '-o', '--output', default='output', metavar='OUTPUT',
        help=('OUTPUT prefix, will write fragmentary file to OUTPUT.frag.fas '
              'and full-length file to OUTPUT.full.fas (default=output)'),
        type=str, dest='output')
    args = parser.parse_args()
    return args


def main():
    args = parse_args()
    sequences = MutableAlignment()
    assert os.path.isfile(args.input) and os.access(args.input, os.R_OK), \
        "Input file %s does not exist\n" % args.input
    sequences.read_file_object(args.input)
    frag = MutableAlignment()
    full = MutableAlignment()

    for (key, seq) in sequences.items():
        if (len(seq) <= args.threshold):
            frag[key] = seq
        else:
            full[key] = seq
    frag.write_to_path("%s.frag.fas" % args.output)
    full.write_to_path("%s.full.fas" % args.output)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-


'''
Created on May 13, 2014
@author: namphuon
'''
import argparse
import os
from argparse import Namespace
import sepp.config
import sepp.metagenomics
import sepp


def parse_args():
    parser = argparse.ArgumentParser(
        description='Performs various tools for TIPP.')
    parser.add_argument(
        '-g', '--gene', default=None, metavar='GENE',
        help='use GENE\'s reference package',
        type=str, dest='gene')
    parser.add_argument(
        '-a', '--action', default=None, metavar='ACTION', help='Run ACTION',
        required=True, type=str, dest='action')
    parser.add_argument(
        '-o', '--output', default='output', metavar='OUTPUT',
        help='OUTPUT directory', type=str, dest='output')
    parser.add_argument(
        '-p', '--prefix', default='prefix', metavar='PREFIX', help='PREFIX',
        type=str, dest='prefix')
    parser.add_argument(
        '-i', '--input', default='input', metavar='INPUT',
        help='INPUT destination',
        type=str, dest='input')
    parser.add_argument(
        '-t', '--threshold', default=0.95, metavar='THRESHOLD',
        help='threshold for classification',
        type=float, dest='threshold')

    args = parser.parse_args()
    return args


root_p = open(os.path.join(os.path.split(
    os.path.split(sepp.__file__)[0])[0], "home.path")).readlines()[0].strip()
tipp_config_path = os.path.join(root_p, "tipp.config")


def profile(input, gene, output, prefix, threshold):
    sepp.config.set_main_config_path(tipp_config_path)
    opts = Namespace()
    sepp.config._read_config_file(open(tipp_config_path, 'r'), opts)
    (taxon_map, level_map, key_map) = sepp.metagenomics.load_taxonomy(
        "%s/refpkg/%s.refpkg/all_taxon.taxonomy" % (opts.reference.path, gene))
    gene_classification = sepp.metagenomics.generate_classification(
        input, threshold)
    sepp.metagenomics.remove_unclassified_level(gene_classification)
    sepp.metagenomics.write_classification(
        gene_classification, "%s/%s.classification" % (output, prefix))
    sepp.metagenomics.write_abundance(gene_classification, output)


def main():
    args = parse_args()
    if (args.action == 'profile'):
        profile(args.input, args.gene, args.output, args.prefix,
                args.threshold)


if __name__ == "__main__":
    main()
#



'''
Created on Aug 22, 2012

This is meant to emulate a Hierarchical SEPP. We have four types of jobs that
emulate SEPP jobs. These are called in parallel using a a DAG that would be
structurally identical to the hierarchical SEPP and can map to that problem
directly.

@author: smirarab
'''
from sepp.scheduler import Job, JobPool, Join
from random import random
import sys
import os
from multiprocessing import Lock, cpu_count
from sepp.problem import Problem
import unittest


# Depth of problem hierarchy (for decomposition). Tips are equivalent
# of alignment subsets
DEPTH = 2
# The level subproblems are aggregated (like placement subsets)
SUMMERIZE_LEVEL = 1


def trimstr(i):
    s = str(i)
    return s if len(s) < 101 else "%s ..." % (s[0:100])


class TestProblem(Problem):
    def __init__(self, parent):
        Problem.__init__(self, parent)
        self.jobs = dict(buildmodel=None, searchfragment=None, applymodel=None,
                         summarize=None)
        self.fragments = []

    def get_node_label(self):
        # str(self.level),str(self.result)
        return "%s" % (str(self.get_job_result_by_name("buildmodel")))


class GenericJob(Job):
    '''
    Base class of all other job types
    '''
    def __init__(self, problem):
        Job.__init__(self)
        problem.add_job(self.type, self)
        self.problem_name = "level " + str(problem.level)
        # self.add_call_Back(lambda res: problem.add_result_to_probl
        # em_object(res, self.type))

    def print_result(self, result):
        print("Process [%s]: %s finished with results: %s" % (
              os.getpid(), self.type, trimstr(result)), file=sys.stderr)


class BuildModelJob(GenericJob):
    '''
    Build a Model (equivalent of HMMERBUILD)
    '''
    def __init__(self, problem):
        self.type = "buildmodel"
        GenericJob.__init__(self, problem)
        self.add_call_Back(self.print_result)
        # self.someclass

    def run(self):
        '''
        The model is going to be a random number calculated in an unnecessarily
        lengthy fashion.
        '''
        print("Process [%s]: buildmodel running %s" % (
            os.getpid(), self.problem_name), file=sys.stderr)
        h = 0
        step = random() / 10000
        for i in range(0, 10000):
            h += step * i
            # time.sleep(step/100)

        return int(h)


class SearchJob(GenericJob):
    def __init__(self, problem):
        '''
        Search fragments against a Model (equivalent of HMMERSEARCH).

        This jobs depend on BuildModel job of the same subproblem, and also
        Search job __parent problems (if any)
        '''
        self.type = "searchfragment"
        GenericJob.__init__(self, problem)
        '''
        Both the model and fragments list are inputs to this jobs. These are
        not calculated yet at the time the job object is *created*. But they
        should be set before the job in *enqueued*. By the time this job is
        enqueued, the dependencies are run, and hence the model is available
        (from BuildModel) and fragments are known (from Search of parents)
        '''
        self.model = None
        self.fragments = None

    def run(self):
        '''
        Simply find the difference between each fragment value and the model
        associated with this job. Do this in an inefficient way.
        '''
        print("Process [%s]: %s running %s with model %d" % (
            os.getpid(), self.type, self.problem_name, self.model),
            file=sys.stderr)
        # time.sleep(random()/10)
        # self.state = step
        for i in range(1, 2000):

            ret = [abs(self.model-fragment) for fragment in self.fragments]
        return ret


class ApplyModelJob(GenericJob):
    def __init__(self, problem):
        '''
        Apply the model to fragments associated with each subproblem
        (equivalent of HMMERALIGN)

        This job is run only on tip subproblems, and depends on the Search
        jobs. Input again is a model and a set of fragments. These are both
        available only after Search job is run. These values should be set
        before an ApplyModel job is enqueued.
        '''
        self.type = "applymodel"
        GenericJob.__init__(self, problem)
        self.fragments = None
        self.model = None

    def run(self):
        '''
        Do some mathematical calculation involving each fragment and the model.
        '''
        print("Process [%s]: %s running %s with model %d" % (
            os.getpid(), self.type, self.problem_name, self.model),
            file=sys.stderr)
        # time.sleep(random()/20)
        # self.state = step
        for i in range(1, 2000):
            ret = [self.model * fragment / 2000.0
                   for fragment in self.fragments]

        return ret


class SummarizeJob(GenericJob):
    def __init__(self, problem):
        '''
        Use model-applied fragment values (equivalent of extended alignments)
        to compute some new results (equivalent of pplacer)

        This jobs
        '''
        self.type = "summarize"
        GenericJob.__init__(self, problem)
        self.resultsPerTipSubproblem = None

    def run(self):
        '''
        Simply normalize the model-applied fragments across
        '''
        print("Process [%s]: %s running %s with results from tips %s" % (
                os.getpid(), self.type, self.problem_name,
                trimstr(self.resultsPerTipSubproblem)), file=sys.stderr)
        # time.sleep(random()/20)
        # self.state = step
        all_fragments = []
        for fragments in self.resultsPerTipSubproblem:
            all_fragments.extend(fragments)
        for i in range(1, 20000):
            fsum = sum(all_fragments)
            ret = [(fragment + .0) / fsum for fragment in all_fragments]
        return ret


class Join_BuildModel_SearchFragment(Join):
    '''
    A join object used to join buildmodel jobs with their __parent
    searchfragments jobs. The input to the constructor is the
    grandparent problem; i.e., the problem two level above buildmodel jobs,
    and one level above searchfragment jobs to be joined together
    '''
    def __init__(self, problem):
        Join.__init__(self)
        self.grandparent_problem = problem
        '''automatically add appropriate jobs to this join'''
        for l1 in self.grandparent_problem.children:
            self.add_job(l1.jobs["searchfragment"])
            for l2 in l1.children:
                self.add_job(l2.jobs["buildmodel"])

    def perform(self):
        print("Process [%s]: Join_BuildModel_SearchFragment joining %s" % (
            os.getpid(), self.grandparent_problem), file=sys.stderr)
        '''
        1 - start from grandparent fragments.
        2 - Based on results from joined search operations, figure out for each
            grandparent fragments whether it is closer to parent1 or parent2
            (grandparent's children), and divide the fragments accordingly
        3 - Set fragments attribute of parent1 and parent2 based on preceding
            calculation
        4 - For each of grandchildrens of the grandparent problem,
            4-1 Set the fragments attribute of their searchfragment job to the
                fragments of its parents (set in step 3)
            4-2 Set the model attribute of their searchfragment job to the
                model computed in its buildmodel job
            4-3 Enqueue its searchfragment job

        The above procedure is equivalent of figuring out the child HMM model
        that a fragment is closer to, and classifying it accordingly.
        '''
        fragments = self.grandparent_problem.fragments
        frags_range = list(range(0, len(fragments)))
        c1_res = self.grandparent_problem.children[0].get_job_result_by_name(
            "searchfragment")
        c2_res = self.grandparent_problem.children[1].get_job_result_by_name(
            "searchfragment")
        model_search = [c1_res[i] - c2_res[i] for i in frags_range]
        self.grandparent_problem.children[0].fragments = [
            fragments[i] for i in frags_range if model_search[i] < 0]
        self.grandparent_problem.children[1].fragments = [
            fragments[i] for i in frags_range if model_search[i] >= 0]
        for l1 in self.grandparent_problem.children:
            fragments = l1.fragments
            for l2 in l1.children:
                l2j = l2.jobs["searchfragment"]
                l2j.model = l2.get_job_result_by_name("buildmodel")
                l2j.fragments = fragments
                JobPool().enqueue_job(l2j)


class Join_ApplyModel_Summarize(Join):
    '''
    This joins applymodel jobs with summirze jobs
    '''
    def __init__(self, problem):
        Join.__init__(self)
        self.summarylevel_problem = problem
        '''Add jobs of applymodel of the grand __parent's children (parents)'''
        for tip in self.summarylevel_problem.iter_leaves():
            self.add_job(tip.jobs["applymodel"])

    def perform(self):
        '''
        Aggregate fragments from tips to the SUMMERIZE_LEVEL level problem,
        and enqueue a summarize job
        '''
        print("Process [%s]: Join_ApplyModel_Summarize joining %s" % (
            os.getpid(), self.summarylevel_problem), file=sys.stderr)
        resultsPerTipSubproblem = []
        for tip in self.summarylevel_problem.iter_leaves():
            resultsPerTipSubproblem.append(tip.get_job_result_by_name(
                "applymodel"))
        self.summarylevel_problem.jobs["summarize"].resultsPerTipSubproblem = \
            resultsPerTipSubproblem
        JobPool().enqueue_job(self.summarylevel_problem.jobs["summarize"])


class Join_tip_searchfragment(Join):
    '''
    After all search jobs have finished on tips, we need to start applying
    the model of each tip subproblem to its fragments. This join takes care of
    that.
    '''
    def __init__(self, root_problem):
        Join.__init__(self)
        self.root_problem = root_problem
        for p in root_problem.iter_leaves():
            self.add_job(p.jobs["searchfragment"])

    def perform(self):
        '''
        First print out some summary of everything up to here.
        Then update applymodel jobs with correct fragment and model, and
        then enqueue them.
        '''
        print("Process [%s]: Join_tip_searchfragment joining %s" % (
            os.getpid(), self.root_problem), file=sys.stderr)

        def print_fragments(problem):
            print(
                "\t" * (problem.level),
                trimstr(problem.get_job_result_by_name("buildmodel")),
                trimstr(problem.fragments))
            for c in problem.children:
                print_fragments(c)
        print_fragments(self.root_problem)

        for p in root_problem.iter_leaves():
            j = p.jobs["applymodel"]
            j.fragments = p.jobs["searchfragment"].fragments
            j.model = p.jobs["searchfragment"].model
            JobPool().enqueue_job(j)

    def __str__(self):
        return "join Join_tip_searchfragment for", self.root_problem


def build_subproblems(problem=None):
    ''' Makes a fully balanced problem (binary) tree upto level DEPTH'''
    if problem is None:
        problem = TestProblem(None)
        ''' Root problem needs to have fragments assigned to it'''
        problem.fragments = [int(random() * 1000000) for i in range(1, 2000)]
    if problem.level < DEPTH:
        subproblem = TestProblem(problem)
        build_subproblems(subproblem)
        problem.add_child(subproblem)
        subproblem = TestProblem(problem)
        build_subproblems(subproblem)
        problem.add_child(subproblem)
        # problem.ad
    return problem


def build_job_dag(root_problem):
    def recursive_add_buildmodel_and_searchfragment_job(problem):
        '''build a search job and a buildmodel job for each problem in
        the problem tree.'''
        if problem.parent is not None:
            BuildModelJob(problem)
            SearchJob(problem)
        for subp in problem.children:
            recursive_add_buildmodel_and_searchfragment_job(subp)

    def connect_buildmodel_and_searchfragment_jobs(problem):
        '''
        Connect all searchfragment jobs for nodes under a certain node
        ("grandparent") with all the buildmodels jobs under all those nodes
        (grandchildren).
        '''
        # and sum((len(c.children) for c in problem.children)) > 0:
        if len(problem.children) > 0:
            Join_BuildModel_SearchFragment(problem)
            for c in problem.children:
                connect_buildmodel_and_searchfragment_jobs(c)

    recursive_add_buildmodel_and_searchfragment_job(root_problem)
    connect_buildmodel_and_searchfragment_jobs(root_problem)

    ''' All the tip search problems should be joined'''
    Join_tip_searchfragment(root_problem)

    ''' only for leaf problems create a apply model job'''
    for problem in root_problem.iter_leaves():
        ApplyModelJob(problem)
    ''' only for problems at SUMMERIZE_LEVEL create a summarize job'''
    for problem in root_problem.iter_nodes_at_level(SUMMERIZE_LEVEL):
        SummarizeJob(problem)
        '''join these summarizejobs with all the leave applymodel
           jobs under them'''
        Join_ApplyModel_Summarize(problem)

    '''
    highest level needs special handling. Make sure you add a call_back
    function to call searchfragment after buildmodel for the top level
    buildmodel jobs. Note that this could have been also achived using a join
    that has only one job.'''
    for c in root_problem.children:
        def enq_job_searchfragment(result, next_job):
            next_job.model = result
            next_job.fragments = root_problem.fragments
            JobPool().enqueue_job(next_job)
        c.jobs["buildmodel"].add_call_Back(
            lambda result, next_job=c.jobs["searchfragment"]:
                enq_job_searchfragment(result, next_job))


s = 0
lock = Lock()
root_problem = None


def run():
    global root_problem
    JobPool().terminate()
    JobPool().__init__(2)
    JobPool(2)

    '''build the problem structure'''
    root_problem = build_subproblems()

    '''build the dat of jobs'''
    build_job_dag(root_problem)

    '''All buildmodel jobs are ready to be started (i.e. no dependency).
    Queue them up. Once they run, they will automatically enqueue the rest of
    the DAG through joins and callbacks '''
    def enqueue_buildmodel_job(problem):
        if problem.parent is not None:
            JobPool().enqueue_job(problem.jobs["buildmodel"])
        for child in problem.children:
            enqueue_buildmodel_job(child)
    enqueue_buildmodel_job(root_problem)

    '''Wait for all jobs to finish'''
    JobPool().wait_for_all_jobs()

    ''' print out the results of summarize jobs. We could merge the
    results from all summarize jobs here if we wanted '''
    for problem in root_problem.iter_nodes_at_level(SUMMERIZE_LEVEL):
        print(trimstr(problem.get_job_result_by_name("buildmodel")),
              trimstr(problem.get_job_result_by_name("summarize")))
    # print [str(x) for x in root_problem.iter_leaves()]


class Test(unittest.TestCase):
    def tearDown(self):
        # clean up JobPool for other unit tests
        JobPool().terminate()
        JobPool().__init__(cpu_count())

    def test_me(self):
        run()


if __name__ == '__main__':
    unittest.main()



'''
Created on Aug 22, 2012

@author: smirarab
'''
from multiprocessing import cpu_count, Pool, Lock
import copy
import traceback
from sepp import get_logger

_LOG = get_logger(__name__)


class Job(object):
    '''
    The base class for a job that can be added to and executed form our
    JobPool. Subclasses of Job need to overwrite the 'run' method to perform
    their tasks. A correct parallel way of running the job is by adding it to
    the JobPool.

    A Job object has  no direct access to its AsyncResult object, and thus has
    no way of checking whether it has finished (for good reasons).
    AsyncResult object of a job, and more generally checking its status, should
    instead occur through the JobPool object. All job management tasks should
    occur through JobPool. job.result should be accessed only once one is sure
    that there the job has finished successfully (through JobPool).

    A Job has a list of callback functions, which are called (in the order
    they were added) when the job finishes running successfully. To address
    some pickling issues, this list is kept as part of the
    JobPool object. It should be noted that a Callback function is expected to
    never raise an exception. If an exception is raised,
    it is printed out. Also, To avoid deadlock, the status of the job is set to
    successful even if callbacks throw an exception. However, the error is
    added to the list of errrors for the jobs.

    The result of the job is simply the return value of the "run" method.
    Return value of the run method gets automatically set to job.result once
    the job finishes running *successfully*. callback functions are also called
    only if Job finishes successfully, and therefore can assume that job.result
    is set. If "run" raises an exception, its job.result is never set, and
    job.result_set is left as False. However, a proper check to see if a job
    has failed has to happen through the JobPool to which it was submitted.

    In general, if the "run" method of a Job modifies the state of the object,
    those changes will NOT be propagated to other processes, including the
    parent process that runs the JobPool. The only thing that gets passed to
    the parent process is the "result" of a job.
    Similarly, if the state of a Job object is modified after it is put in the
    queue, the process that runs the Job might not see those changes.
    For these reasons, it is not wise to use Job class attributes for saving
    state within the run method.

    Callback functions can access global state, but they should do so in a
    thread-safe fashion, because other job's callback functions could be trying
    to access the same global state at the same time, giving rise to race
    conditions. Usage of a Lock, a Queue, or Pipe is recommended.

    In cases when multiple jobs need to finish before another task can start
    the concept of a Join (see below) is useful.
    '''

    def __init__(self):
        '''
        Constructor
        '''
        self.result = None
        self.result_set = False
        self.errors = []
        self.ignore_error = False

    def __call__(self):
        ''' This method makes this class a callable object'''
        return self.run()

    def run(self):
        '''
        This is the method that does the actual job. It should be overridden
        in subclasses. The return value is what gets set as self.result, and
        is passed from child process to the parent process (and should be
        therefore picklable)
        '''
        raise NotImplementedError

    def add_call_Back(self, function):
        ''' Adds a callback function to this job. The callback function should
        accept one parameter, which is going to be set to self.result.

        Callback functions should be thread-safe. If the access global state,
        they should do with care. They should also return rather quickly, as
        they can block others.

        It is safe (to the best of our effort/knowledge) for a callback
        function to enqueue more jobs.
        '''
        if not hasattr(function, "__call__"):
            raise TypeError(
                "Expecting a callable object. Instead getting a %s" %
                type(function))
        JobPool()._add_callback_for_job(self, function)

    def _finished(self, results):
        '''
        This is a default callback function automatically added for all jobs.
        It is the first function called when job finishes, and simply sets
        the result of the job.
        '''
        self.result = results
        self.result_set = True


def check_object(job):
    ''' Check to make sure an object is a legitimate Job object'''
    if not isinstance(job, Job):
        raise TypeError("Expecting a Job object, instead getting a %s" %
                        type(job))
    pass


class Join(object):
    '''
    This class allows synchronization between multiple jobs. If a collection of
    jobs needs to finish before we can perform another task (such as enqueuing
    more jobs), using callbacks is not sufficient. A Join is helpful in these
    situations.

    A join is basically two things. A set of Jobs that are joined together,
    and a perform() method that does what the Join needs to do. Subclasses of
    this abstract Join class need to override perform() and do what they need
    to do.

    A typical use of a Join consists of initializing a new Join object and
    simply adding jobs to it (one by one). There is no need to "register" the
    join with the schedule, as that step is done automatically when jobs are
    added to the join.

    When all the jobs added to a join have finished running, the perform()
    method of the join is automatically called by the scheduler.

    Note that it makes sense to build a join and add jobs to it *before* those
    jobs are enqueued in the pool. The Join may or may not work properly
    otherwise (this could be better tested and improved).

    A typical perform method would read results from the joined jobs, would
    process the results, prepare inputs for a bunch of downstream jobs,
    would set those inputs as attributes of those downstream jobs, and would
    finally enqueue them.

    It is important to note that a Join's perform method is called basically
    as the last callback function of the last job in the join. As such,
    1) it is run on the main process
    2) when it is performed, all the jobs have finished, and their results are
       set in the job objects, but the last job that finished does not yet have
       its status set to ready. This is crucial because otherwise the link
       could break (i.e. if a wait_for_all_job gets unblocked because there is
       no more jobs in the queue before this job is run)
    3) It needs to be relatively fast. In particular, it should not perform
       heavy duty tasks, and it should not block or wait on other events.
    '''
    def __init__(self):
        self._jobs = set()
        self._lock = Lock()
        self._joined = False

    def perform(self):
        '''
        Perform the main task of the Join. This method needs to be implemented
        by subclasses.
        '''
        raise NotImplementedError

    def depends_on(self, job):
        '''returns true if this join depends on a given job; false
           otherwise'''
        check_object(job)
        self._lock.acquire()
        ret = job in self._jobs
        self._lock.release()
        return ret

    # internal method for checking join status.
    def _assert_has_not_finished(self):
        if self._joined:
            raise Exception(
                "Adding a job to a Join object that has already joined.")

    def add_job(self, job):
        '''
        Add a new job to this join. The job is ideally not queued yet.
        '''
        check_object(job)
        self._lock.acquire()
        self._assert_has_not_finished()
        self._jobs.add(job)
        JobPool()._add_jobs_to_join([job], self)
        self._lock.release()
        self._tick(None)

#    def add_jobs(self, jobs):
#        for job in jobs: check_object(job)
#        self._lock.acquire()
#        self._assert_has_not_finished()
#        self._jobs.update(set(jobs))
#        JobPool()._add_jobs_to_join(job, self)
#        self._lock.release()

#    def replace_job(self, old_job, new_job):
#        check_object(new_job)
#        self._lock.acquire()
#        self._assert_has_not_finished()
#        if not self.depends_on(old_job):
#            self._lock.release()
#            raise KeyError("the job to be replaced is not part of this
#                            join: %s" %str(old_job))
#        self._jobs.remove(old_job)
#        self._jobs.add(new_job)
#        JobPool()._del_jobs_from_join([old_job])
#        JobPool()._add_jobs_to_join([new_job], self)
#        self._lock.release()

    def _tick(self, ticking_job):
        # internal method for synchronizing with JobPool. Every time a
        # job finishes running, the Join is ticked. If all jobs except the
        # ticking job have finished running, the join is "performed".
        # Note that when join is performed, the status of one of the jobs
        # (ticking job = the last job that finished running) could still be
        # "not-ready",
        # since ticks happens as part of last jobs's callbacks.
        # This is necessary; otherwise, wait_for_all_jobs could finish without
        # waiting for new jobs to be enqueued from the join(). This behavior
        # ensures that a DAG is never broken without enqueuing subsequent
        # parts.
        if self._joined:
            # TODO: should this be an error?
            return
        pool = JobPool()
        self._lock.acquire()
        for job in self._jobs:
            if (job != ticking_job) and (
                (not pool.is_job_queued(job)) or pool.is_job_running(job)
               ):
                self._lock.release()
                return
        self.perform()

        self._joined = True
        self._lock.release()


default_cpus = None


'''
JobPool is provided as a singleton class
'''
_jobPool = None


def JobPool(cpus=None):
    ''' The JobPool singelton instance is returned.
    If the function is called for the first time, a new instance is created
    with the given number of cpus. Subsequent calls return the existing
    instances, if the number of requested cpus match the existing instance, or
    is None. If the number of requested cpus in subsequent calls is different
    from that of the original call, an error is raised.
    '''
    global _jobPool
    global default_cpus
    if cpus is None:
        cpus = default_cpus
    if _jobPool is None:
        _jobPool = _JobPool(cpus)
    if cpus is not None and cpus != _jobPool.cpus:
        raise Exception(
            ("An instance of Singleton class JobPool with %d cpus exists. "
             "You requested %d cpus")
            % (_jobPool.cpus, cpus))
    return _jobPool


class JobError(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)


class _JobPool:
    '''
    A singelton class, managing all Jobs.
    Use JobPool() function to get the instance.
    Can be initialized only once, with a given number of cpus.
    If number of cpus is left blank, all but two cpus (or only 1 cpu if less
    than three available) are used Jobs can be queued for execution. When a
    job finishes, its callback functions are called.
    One can choose to wait for all jobs in the queue to stop.
    Scheduling and managing jobs
    '''

    def __init__(self, cpunum):
        self.cpus = cpunum if cpunum is not None else cpu_count()
        self._pool = Pool(self.cpus)
        self._async_restults = {}
        self._callBack_lists = {}
        self._joins = {}
        self._lock = Lock()

    def enqueue_job(self, job):
        check_object(job)

        def call_back(job, result, callBackCopy):
            try:
                job._finished(result)
                '''Then run all the callback in order, passing result
                   as parameter'''
                for callback in callBackCopy:
                    callback(*[result])
                self._lock.acquire()
                jobJoins = copy.copy(self._joins.get(job, []))
                self._lock.release()
                for join in jobJoins:
                    # print "ticking, ", join
                    join._tick(job)
            except Exception as e:
                # TODO: currently callback exceptions are simply ignored.
                # Is there a better solution?
                job.errors.append(e)
                traceback.print_exc()

        ''' We need to backup callbacks of a Job.
        The following line ensures that pickling issues do not arise'''
        callBackCopy = self._callBack_lists.get(job, [])
        ''' Add jobs to the _pool, and save resulting AsyncResult function in a
        safe manner'''
        self._lock.acquire()
        result = self._pool.apply_async(
            job, callback=lambda res: call_back(job, res, callBackCopy))
        self._async_restults[job] = result
        self._lock.release()

    def _add_callback_for_job(self, job, callback):
        self._lock.acquire()
        if job not in self._callBack_lists:
            self._callBack_lists[job] = []
        self._callBack_lists[job].append(callback)
        self._lock.release()

    def wait_for_all_jobs(self, ignore_error=False):
        '''
        waits for all jobs in the queue to finish running. Makes attempt to
        wait for newly added jobs as well.

        If ignore_error is false, this method stops after the first job with an
        error and with job.ignore_error == True is discovered, and raises
        the error of that job.

        This method returns whether all finished jobs have finished
        successfully. So, if some of the jobs have ignore_error set, or if
        ignore_error parameter is set, this method does not raise those
        exceptions, but still the return value tells you whether there are any
        jobs with error.
        '''
        more = True
        while more:
            for job, result in list(self._async_restults.items()):
                try:
                    result.get()
                    if len(job.errors) != 0:
                        raise Exception(job.errors[0])
                except Exception:
                    if not ignore_error and not job.ignore_error:
                        raise
            ''' _lock is acquired so that while we are iterating through the
            list no other job can add to the list (i.e. from their
            callbacks)'''
            self._lock.acquire()
            more = False
            hasFailed = False
            for (job, result) in self._async_restults.items():
                if not result.ready():
                    more = True
                elif (not result.successful()) or (len(job.errors) != 0):
                    if not ignore_error and not job.ignore_error:
                        result.get()
                    hasFailed = True
            self._lock.release()

        return not hasFailed

    def is_job_running(self, job):
        '''Check to see if a given job has finished running'''
        return not self.get_asynch_result_object(job).ready()

    def is_job_queued(self, job):
        return job in self._async_restults

    def get_asynch_result_object(self, job):
        ''' returns an instance of multiprocessing.pool.AsyncResult class,
        corresponding to the execution of the given job. This object can
        be used to check status of a job'''
        self._lock.acquire()
        if job not in self._async_restults.keys():
            raise KeyError(
                "Job %s not in the queue, or _pool terminated." % str(job))
        resobj = self._async_restults[job]
        self._lock.release()
        return resobj

    def _is_job_failed(self, job, res):
        return (res.ready() and not res.successful()) or (len(job.errors) != 0)

    def get_failed_jobs(self):
        ''' Returns a list of files that have failed (i.e. their run method has
        finished with an exception raised)'''
        ret = []
        self._lock.acquire()
        for job, res in self._async_restults.items():
            if self._is_job_failed(job, res):
                ret.append(job)
        self._lock.release()
        return ret

    def get_job_error(self, job):
        '''
        If the given job has finished successfully, this method simply returns
        None. Otherwise, it returns the exception that terminated the given job
        '''
        res = self.get_asynch_result_object(job)
        if not self._is_job_failed(job, res):
            return None
        elif len(job.errors) != 0:
            return job.errors
        else:
            try:
                res.get()
            except Exception as e:
                return e

    def get_all_job_errors(self):
        ''' Returns a list of all exceptions raised in all jobs executed in
        this _pool. An empty list means no exceptions '''
        self._lock.acquire()
        jobs = set(self._async_restults.keys())
        self._lock.release()
        ret = [x for x in
               [self.get_job_error(job) for job in jobs]
               if x is not None]
        return ret

    def _add_jobs_to_join(self, jobs, join):
        self._lock.acquire()
        for job in jobs:
            self._joins[job] = self._joins.get(job, []) + [join]
        self._lock.release()

    def _del_jobs_from_join(self, jobs, join):
        self._lock.acquire()
        for job in jobs:
            self._joins.get(job).remove(job)
        self._lock.release()

    def terminate(self):
        ''' Terminate all jobs in a queue'''
        self._lock.acquire()
        self._pool.close()
        self._pool.terminate()
        self._pool = None
        self._async_restults = {}
        self._lock.release()



'''
Created on Nov 20, 2012

@author: smirarab
'''
import unittest
from sepp.jobs import MergeJsonJob
from sepp.scheduler import JobPool
from multiprocessing import cpu_count
import sys
import os.path
from sepp.filemgr import get_data_path
import sepp.config
from argparse import Namespace


class Test(unittest.TestCase):
    def tearDown(self):
        # clean up JobPool for other unit tests
        JobPool().terminate()
        JobPool().__init__(cpu_count())

    def testMerge(self):
        sys.argv = [sys.argv[0]]
        stdindata = open(get_data_path("tmp/tempmerge")).read()

        # write path of seppJsonMerger.jar into configuration
        setattr(sepp.config.options(), 'jsonmerger', Namespace(
            path=get_data_path("../../../tools/merge/seppJsonMerger.jar")))
        mergeJsonJob = MergeJsonJob()
        mergeJsonJob.setup(stdindata.replace(
            'data/tmp/pplacer.extended',
            os.path.dirname(get_data_path("tmp/tempmerge")) +
            '/pplacer.extended'), get_data_path("tmp/mergedfile"))
        mergeJsonJob.run()


if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testMerge']
    unittest.main()



'''
Created on Aug 22, 2012

@author: smirarab
'''
from sepp.scheduler import Job, JobPool
from multiprocessing import cpu_count
from random import random
from multiprocessing import Lock
import time
import unittest


class TestJob(Job):
    def __init__(self, jobname):
        Job.__init__(self)
        global s
        lock.acquire()
        s += 1
        lock.release()
        self.jobname = jobname
        self.state = None

        def add_a_child(parent):
            # print("Adding a child job for %s" % (parent), file=sys.stderr)
            JobPool().enqueue_job(TestJob("%s.child" % parent))
            # print "Added a child for: ",parent

        def a_very_bad_callback():
            raise Exception("A bad callback raises an exception.")

        ''' Make a random number of these jobs, child-bearing'''
        if random() < 0.75:
            self.add_call_Back(
                lambda result, parent=self.jobname: add_a_child(parent))

        if random() < 0.1:
            self.add_call_Back(lambda result: a_very_bad_callback())

        self.add_call_Back(self.print_res)

    def print_res(self, result):
        global s
        # print("%s returned: %d" % (self.jobname, self.result))
        lock.acquire()
        s -= 1
        lock.release()

    def run(self):
        # print("Process [%s]: running %s" % (os.getpid(), self.jobname),
        #       file=sys.stderr)
        # Adding the following line results in a failure:
        # AssertionError: daemonic processes are not allowed to have children.
        # This is expected because the new process is going to try to start a
        # new JobPool, which is not allowed.
        # print "cpus ",JobPool().cpus
        h = 0
        step = random()
        if step < 0.1:
            raise Exception("Some (truly) random error occurred in job %s." %
                            self.jobname)
        for i in range(0, 100):
            h += step * i
            time.sleep(step / 100)
        # self.state = step
        return h


s = 0
lock = Lock()


def run():
    global pool
    pool1 = JobPool(2)
    pool2 = JobPool()
    if pool1 != pool2:
        raise Exception("hmmm, I thought JobPool is 'Singleton'")
    try:
        JobPool(4)
    except Exception as e:
        print(("As expected, making a new JobPool with a"
               " different cpu count failed: %s") % e)

    pool = JobPool()
    jobs = []
    for j in range(1, 20):
        job = TestJob(str(j))
        jobs.append(job)
        pool.enqueue_job(job)

    sample_job = pool.get_asynch_result_object(jobs[3])

    # pool.terminate()

    pool.wait_for_all_jobs(ignore_error=True)

    # Test one of the jobs, to see if it is successful
    if sample_job.ready() and sample_job.successful():
        assert(jobs[3].result_set is True)
    else:
        assert(jobs[3].result_set is False)

    errors = pool.get_all_job_errors()
    # print("Following job errors were raised:", errors)

    try:
        pool.wait_for_all_jobs(ignore_error=False)
    except Exception as e:
        print("Seems we have some jobs that failed (expected): ", e)

    errs = [pool.get_job_error(job) for job in pool.get_failed_jobs()]

    # print(errs)

    assert len(errs) == len(errors), \
        "Number of errors from failed jobs: %d. Number of errors: %d" % \
        (len(errs), len(errors))
    assert False not in [x in errors for x in errs]

    # print [job.state for job in jobs]
    # print("Number of started jobs - number of printed results:", s)
    # print("Number of failed jobs:", len(errors))
    assert s == len(errors), \
        "Parallelization Error, what happened to the rest?"

    # print("Everything seems fine!")


class Test(unittest.TestCase):
    def tearDown(self):
        # clean up JobPool for other unit tests
        JobPool().terminate()
        JobPool().__init__(cpu_count())

    def test_me(self):
        run()


if __name__ == '__main__':
    unittest.main()



'''
Collection of functions for metagenomic pipeline for taxonomic classification
Created on June 3, 2014

@author: namphuon
'''

from sepp.scheduler import Join, JobPool
from sepp.problem import SeppProblem
from sepp.jobs import HMMAlignJob
from sepp import get_logger
_LOG = get_logger(__name__)


class JoinBlastJobs(Join):
    '''
    After all blast search jobs have finished on markers, we need to figure out
    which fragment goes to which marker.
    This join takes care of that step.
    '''
    def __init__(self):
        Join.__init__(self)

    def setup_with_root_problem(self, root_problem):
        self.root_problem = root_problem
        for p in root_problem.iter_leaves():
            self.add_job(p.jobs["blastsearch"])

    def figureout_fragment_marker(self):
        ''' Figure out which fragment should go to which marker'''
        if "fragments.distribution.done" in self.root_problem.annotations:
            return
        max_evalues = dict([(name, (None, None))
                            for name in self.root_problem.fragments.keys()])
        for fragment_chunk_problem in self.root_problem.iter_leaves():
            align_problem = fragment_chunk_problem.get_parent()
            assert isinstance(align_problem, SeppProblem)
            '''For each subproblem start with an empty set of fragments,
            and add to them as we encounter new best hits for that
            subproblem'''
            if align_problem.fragments is None:
                align_problem.fragments = \
                    self.root_problem.fragments.get_soft_sub_alignment([])
            search_res = fragment_chunk_problem.get_job_result_by_name(
                "hmmsearch")
            for key in search_res.keys():
                (best_value, prev_align_problem) = max_evalues[key]
                ''' If this is better than previous best hit, remove this
                fragment from the previous hit, and add it to this subproblem
                '''
                if best_value is None or (best_value < search_res[key][1]):
                    max_evalues[key] = (search_res[key][1], align_problem)

        # TODO: is the following efficient enough? Do we need to make lists
        # and then turn them to sets?
        notScored = []
        for key, v in max_evalues.items():
            if v[1] is None:
                notScored.append(key)
            else:
                v[1].fragments.seq_names.add(key)

        self.root_problem.annotations["fragments.distribution.done"] = 1

        ''' Make sure all fragments are in at least one subproblem.
        TODO: what to do with those that are not?  For now, only output
        warning message'''
        # notScored = [k for k, v in max_evalues.iteritems() if v[1] is None]
        _LOG.warning(
            "Fragments %s are not scored against any subset" % str(notScored))
        # assert len(notScored) == 0, "Fragments %s are not scored against any
        # subset" %str(notScored)

    def perform(self):
        '''
        Distributes fragments to alignments subsets with best score,
        and runs align jobs on those. Also, creates new chunks of fragments
        for better parallelism.
        '''

        ''' Figure out which fragment should go to which subproblem'''
        self.figureout_fragment_subset()

        ''' For each alignment subproblem,
        1) make sure its fragments are evenly distributed to fragment chunks.
        2) Setup alignment jobs for its children and enqueue them'''
        alg_problems = [alg for p in self.root_problem.children
                        for alg in p.children]
        for alg_problem in alg_problems:
            assert isinstance(alg_problem, SeppProblem)
            chunks = len(alg_problem.get_children())
            fragment_chunks = alg_problem.fragments.divide_to_equal_chunks(
                chunks)

            ''' Now setup alignment jobs and enqueue them'''
            for (i, fragment_chunk_problem) in enumerate(alg_problem.children):
                fragment_chunk_problem.fragments = fragment_chunks[i]
                aj = fragment_chunk_problem.jobs['hmmalign']
                assert isinstance(aj, HMMAlignJob)
                ''' First Complete setting up alignments'''
                aj.hmmmodel = alg_problem.get_job_result_by_name('hmmbuild')
                aj.base_alignment = alg_problem.jobs["hmmbuild"].infile

                if not fragment_chunk_problem.fragments.is_empty():
                    fragment_chunk_problem.fragments.write_to_path(
                        aj.fragments)
                else:
                    aj.fake_run = True
                ''' Now the align job can be put on the queue '''
                JobPool().enqueue_job(aj)

    def __str__(self):
        return "join search jobs for all tips of ", self.root_problem


def blast_fragments(config, input, output):
    '''Blast the fragments against all marker genes+16S sequences'''


def fix_direction(config, input, output):
    '''Fixes the direction of all the reads by searching the
       sequences against each HMM and finding out which one works the best'''


def reverse_sequence(sequence):
    '''Reverse a sequence to be in the same direction as marker sequences
    '''


def read_blast_results(input):
    '''Reads the results for blast
    '''

    # def read_results(self):
    #     '''
    #     Read the Sate log file and get the alignment and tree from file,
    #     copy to output directory
    #     '''
    #     assert os.path.exists('%s/sateout/satejob.out.txt' %
    #                           sepp.filemgr.get_root_temp_dir())
    #     assert os.stat('%s/sateout/satejob.out.txt' %
    #                    sepp.filemgr.get_root_temp_dir())[stat.ST_SIZE] != 0
    #     outfile = open('%s/sateout/satejob.out.txt' %
    #                    sepp.filemgr.get_root_temp_dir(), 'r');
    #     alignment_pattern = re.compile('Writing resulting alignment to (.*)')
    #     tree_pattern = re.compile('Writing resulting tree to (.*)')
    #     tree_file = ''
    #     alignment_file = ''
    #     for line in outfile:
    #         line = line.strip()
    #         result = alignment_pattern.findall(line)
    #         if (len(result) != 0):
    #             alignment_file = result[0]
    #         result = tree_pattern.findall(line)
    #         if (len(result) != 0):
    #             tree_file = result[0]
    #     shutil.copyfile(tree_file, "%s/sate.fasttree" % self.output)
    #     shutil.copyfile(alignment_file, "%s/sate.fasta" % self.output)
    #     return (tree_file,alignment_file)



#!/bin/env python

import sys

raxmlinfo = '''
Base frequencies: %f %f %f %f

Inference[0]: Time %f CAT-based likelihood -0000, best rearrangement setting 5
alpha[0]: 1.000000 rates[0] ac ag at cg ct gt: %f %f %f %f %f %f


NOT conducting any final model optimizations on all 1 trees under CAT-based
model ....

Final GAMMA  likelihood: %f
'''

ls = open(sys.argv[1], 'r').readlines()

for (i, l) in enumerate(ls):
    if "called as follows" in l:
        break
i += 3

v = dict()

for l in ls[i:-1]:
    lr = l[0:-1].split(' ')
    if "<->" in l:
        v[' '.join(lr[1:4])] = float(lr[4])
    elif "freq" in l:
        v[lr[1]] = float(lr[2])
    elif "Final GAMMA" in l and "likelihood" in l:
        v['lk'] = float(lr[-1])
    elif "Overall Time for Tree Evaluation" in l:
        v['tm'] = float(lr[-1])

print('This is a RAxML_info file from a -f e run, automatically reformatted')
print(''.join(ls[0:i]))
print(raxmlinfo % (v['pi(A):'], v['pi(C):'], v['pi(G):'], v['pi(T):'],
                   v['tm'],
                   v['A <-> C:'], v['A <-> G:'], v['A <-> T:'], v['C <-> G:'],
                   v['C <-> T:'], v['G <-> T:'],
                   v['lk']))



# uym2 added
# June 2017
# utils for tree decomposition


from dendropy import Tree
try:
    from queue import Queue  # python 3
except ImportError:
    from Queue import Queue  # python 2
# from tree import PhylogeneticTree
from sepp import get_logger

_LOG = get_logger(__name__)


def decompose_by_diameter(a_tree, strategy, max_size=None, min_size=None,
                          max_diam=None):
    def __ini_record__():
        for node in a_tree.postorder_node_iter():
            __updateNode__(node)

    def __find_midpoint_edge__(t):
        u = t.seed_node.bestLCA.anchor
        uel = u.edge_length if u.edge_length else 0
        d = 0
        while (d+uel < t.seed_node.diameter/2):
            d += uel
            u = u.parent_node
            uel = u.edge_length if u.edge_length else 0
        return u.edge

    def __find_centroid_edge__(t):
        u = t.seed_node
        product = 0
        acc_nleaf = 0

        while not u.is_leaf():
            max_child = None
            max_child_nleaf = 0
            for ch in u.child_node_iter():
                if ch.nleaf > max_child_nleaf:
                    max_child_nleaf = ch.nleaf
                    max_child = ch
            acc_nleaf += (u.nleaf-max_child.nleaf)
            new_product = max_child.nleaf * acc_nleaf
            if new_product <= product:
                break
            product = new_product
            u = max_child

        return u.edge

    def __bisect__(t, e):
        # e = __find_centroid_edge__(t)

        u = e.tail_node
        v = e.head_node

        u.remove_child(v)
        t1 = Tree(seed_node=v)

        if u.num_child_nodes() == 1:
            p = u.parent_node
            v = u.child_nodes()[0]
            l_v = v.edge_length if v.edge_length else 0
            u.remove_child(v)
            # u is the seed_node; this means the tree runs out of all but one
            # side
            if p is None:
                t.seed_node = v
                return t, t1
            l_u = u.edge_length if u.edge_length else 0
            p.remove_child(u)
            p.add_child(v)
            v.edge_length = l_u + l_v
            u = p

        while u is not None:
            __updateNode__(u)
            u = u.parent_node

        return t, t1

    def __clean_up__(t):
        for node in t.postorder_node_iter():
            delattr(node, "nleaf")
            delattr(node, "anchor")
            # delattr(node,"maxheight")
            delattr(node, "maxdepth")
            delattr(node, "diameter")
            # delattr(node,"topo_diam")
            delattr(node, "bestLCA")

    def __updateNode__(node):
        if node.is_leaf():
            node.anchor = node
            # node.maxheight = 0
            node.maxdepth = 0
            node.diameter = 0
            # node.topo_diam = 0
            node.bestLCA = node
            node.nleaf = 1
            return

        # n1 = -1
        # n2 = -1
        d1 = -1
        d2 = -1
        anchor1 = None
        node.diameter = 0
        # node.topo_diam = 0
        node.bestLCA = None
        node.nleaf = 0

        for ch in node.child_node_iter():
            node.nleaf += ch.nleaf
#               n = ch.maxheight + 1
            d = ch.maxdepth + ch.edge_length if ch.edge_length else 0
#               if n > n1:
#                   n2 = n1
#                   n1 = n
#                   anchor2 = anchor1
#                   anchor1 = ch.anchor
#               elif n > n2:
#                   n2 = n
#                   anchor2 = ch.anchor
            if d > d1:
                d2 = d1
                d1 = d
                anchor1 = ch.anchor
            elif d > d2:
                d2 = d
            if ch.diameter > node.diameter:
                node.diameter = ch.diameter
                node.bestLCA = ch.bestLCA
#               node.diameter = max(ch.diameter,node.diameter)

#        node.diameter = max(d1+d2, node.diameter)
        node.maxdepth = d1
#        node.maxheight = n1
        node.anchor = anchor1
        if d1+d2 > node.diameter:
            node.diameter = d1+d2
            node.bestLCA = node

    def __get_breaking_edge__(t, edge_type):
        if t.seed_node.nleaf <= max_size and t.seed_node.diameter <= max_diam:
            return None
        if edge_type == 'midpoint':
            e = __find_midpoint_edge__(t)
        elif edge_type == 'centroid':
            e = __find_centroid_edge__(t)
        else:
            _LOG.warning(("Invalid decomposition type! Please use either "
                          "'midpoint' or 'centroid'"))
            return None

        n = e.head_node.nleaf
        if (n < min_size) or (t.seed_node.nleaf - n) < min_size:
            return None
        return e

    def __check_stop__(t):
        return ((t.seed_node.nleaf <= max_size and
                 t.seed_node.diameter <= max_diam) or
                (t.seed_node.nleaf//2 < min_size))

    def __break_by_MP_centroid__(t):
        e = __get_breaking_edge__(t, 'midpoint')
        if e is None:
            # print("Midpoint failed. Trying centroid decomposition...")
            e = __get_breaking_edge__(t, 'centroid')
        # else:
        #    print("Successfully splitted by midpoint")
        return e

    def __break(t):
        if strategy == "centroid":
            return __get_breaking_edge__(t, 'centroid')
        elif strategy == "midpoint":
            return __break_by_MP_centroid__(t)
        else:
            raise Exception("strategy not valid: %s" % strategy)

    tqueue = Queue()

    _LOG.debug("Starting brlen decomposition ...")
    __ini_record__()
    min_size = min_size if min_size else 0
    max_size = max_size if max_size else a_tree.seed_node.nleaf
    max_diam = max_diam if max_diam else a_tree.seed_node.diameter

    _LOG.debug(
        "Now breaking by %s with min %d and max %d sizes and diameter %f ..." %
        (strategy, min_size, max_size, max_diam))
    # try using midpoint
    e = __break(a_tree)

    if e is None:
        __clean_up__(a_tree)
        return [a_tree]

    treeMap = []
    tqueue.put((a_tree, e))
    while not tqueue.empty():
        t, e = tqueue.get()
        t1, t2 = __bisect__(t, e)
        e1 = __break(t1)
        if e1 is None:
            __clean_up__(t1)
            treeMap.append(t1)
        else:
            tqueue.put((t1, e1))
        e2 = __break(t2)
        if e2 is None:
            __clean_up__(t2)
            treeMap.append(t2)
        else:
            tqueue.put((t2, e2))

    return treeMap



'''
Created on Sep 12, 2012

@author: smirarab
'''
from .scheduler import Job
from sepp.alignment import ReadonlySubalignment, ExtendedAlignment
from sepp import get_logger

_LOG = get_logger(__name__)


class Problem(object):
    '''
    This class gives a hierarchy of Problems. Each problem includes a list
    of subproblems, and a pointer to the __parent problem. Root problem has no
    __parent problem, and tips have empty lists as their __children.

    This class provides the data structure required to decompose a problem
    recursively into subproblems.

    A problem can have many Jobs associated with it. These job objects are kept
    in a dictionary.
    '''
    def __init__(self, parent):
        self.__children = []
        self.__parent = parent
        if parent:
            if not isinstance(parent, Problem):
                raise TypeError("Expecting Problem but founding %s instead" %
                                type(parent))
            parent.add_child(self)
        '''
        This dictionary refers to jobs associated with this subproblem,
        using a name.
        '''
        self.jobs = dict()
        self.annotations = {}

    def get_parent(self):
        return self.__parent
    parent = property(get_parent)

    def get_level(self):
        return 0 if self.parent is None else (self.parent.level + 1)
    level = property(get_level)

    def get_children(self):
        return self.__children
    children = property(get_children)

    def add_child(self, subproblem):
        '''
        Add a child job.
        '''
        self.__children.append(subproblem)

    def iter_leaves(self):
        '''
        Returns an iterator over tips of the problem hierarchy below self.
        '''
        if len(self.__children) == 0:
            yield self
        else:
            for c in self.__children:
                for tip in c.iter_leaves():
                    yield tip

    def iter_nodes_at_level(self, level):
        '''
        a generator function, returning all Problems under the current problem,
        with a label tag equal to the given level.

        Note that this is NOT returning nodes that are 'level' levels below
        self.
        '''
        if self.level == level:
            yield self
        else:
            for c in self.__children:
                for tip in c.iter_nodes_at_level(level):
                    yield tip

    def add_job(self, jobname, job):
        '''
        A a new job to this problem, to be saved with a name
        '''
        self.jobs[jobname] = job

    def get_job_result_by_name(self, jobname, ignore_missing_results=False):
        '''
        returns results of a job, given the job name
        '''
        job = self.jobs[jobname]
        if job is None:
            return None
        assert isinstance(job, Job), "job '%s' is not a Job object" % str(job)
        if job.result_set:
            if hasattr(job, 'results_on_temp') and job.results_on_temp:
                with open(job.result, 'r') as f:
                    res = eval(f.read())
                with open(job.result, 'w') as f:
                    pass
                return res
            else:
                return job.result
        else:
            raise Exception("job '%s' has no results set yet: %s" %
                            (jobname, str(job)))

    def get_node_label(self):
        '''
         returns a label for this Problem. Used in __str__ function.
        '''
        return None

    def get_path_to_root(self):
        ret = [self]
        n = self
        while n.parent is not None:
            ret.append(n.parent)
            n = n.parent
        return ret

    def __str__(self, *args, **kwargs):
        me = self.get_node_label()
        if len(self.__children) == 0:
            return me
        return "(%s)%s" % (
            ",".join((str(child) for child in self.__children)), me)


class SeppProblem(Problem):
    ''' A typical Sepp subproblem, defined by a set of taxa, and a parent

        Adds few basic attributes of a Sepp Problem to the Problem class:
        1- a list of taxa associated with this sub problem
        2- a Readonly subalignment of the parent problem's alginemnt, induced
           by this probelms's taxa (with all gaps columns included)
        3- the subtree of parent problem's tree, induced by this problem's
            taxa. This is a deep copy.
        4- a set of fragments associated with this subproblem

        List of taxa is a required field, provided upon initiailzation.
        Subalignment is automatically computed, and cannot be changed.
        Subtree is automatically computed, but can be overwriteen by the user
        Fragments is not automatically computed, and needs to be set by the
        user
    '''
    def __init__(self, taxa, parent=None):
        Problem.__init__(self, parent)
        self.__taxa = taxa
        self.__subalignment = None
        self.__subtree = None
        self.fragments = None
        self.label = None

    def get_taxa(self):
        return self.__taxa
    taxa = property(get_taxa)

    def get_subalignment(self):
        '''
        If subalignment is not set before, automatically builds a readonly
        subalignment for this subproblem, based on the taxa assigned to this
        subproblem. Otherwise, returns the saves subalignment.
        '''
        if self.__subalignment is None:
            if isinstance(self.parent, SeppProblem) or hasattr(
                self.parent, "subalignment"
               ):
                self.__subalignment = ReadonlySubalignment(
                    self.taxa, self.parent.subalignment)
                # print sorted(self.__subalignment.keys())
                # print sorted(self.parent.subalignment.keys())
                # print "..... %s %s" %(self.label, self.parent.label)
            else:
                raise TypeError(
                    "parent object %s has no subalignment attribute" %
                    self.parent)
        return self.__subalignment

    def set_subalignment(self, alignment):
        self.__subalignment = alignment

    subalignment = property(get_subalignment, set_subalignment)

    def get_subtree(self):
        '''
        If subtree is not assigned before, automatically buils a subtree based
        on the taxa assigned to this subproblem. Otherwise, returns the saves
        subtree.
        '''
        if self.__subtree is None:
            if isinstance(self.parent, SeppProblem) or hasattr(
                self.parent, "subtree"
               ):
                self.__subtree = self.parent.subtree.get_subtree(self.taxa)
            else:
                raise TypeError("parent object %s has no subtree attribute"
                                % self.parent)
        return self.__subtree

    def set_subtree(self, subtree):
        self.__subtree = subtree

    subtree = property(get_subtree, set_subtree)

    def get_node_label(self):
        return self.label

    def write_subalignment_without_allgap_columns(self, path):
        '''
        Writes out the subalignment associated with this subproblem to a file,
        removing all gap columns, but also keeping track of what was removed
        and what was kept. This method keeps track of column names that were
        actually written to file, so that later on, column names could be set
        to the original value. This is crucial for a correct merge.
        (see read_extendend_alignment_and_relabel_columns)
        '''
        mut_subalg = self.subalignment.get_mutable_alignment()
        remaining_cols = mut_subalg.delete_all_gap()
        mut_subalg.write_to_path(path)
        self.annotations["ref.alignment.columns"] = remaining_cols
        _LOG.debug(
            ("subalignment without allgap columns written to %s; "
             "%d columns remaining.") % (path, len(remaining_cols)))
        return remaining_cols

    def read_extendend_alignment_and_relabel_columns(
            self, orig_path, extension_path, convert_to_string=True):
        '''
        This method goes with write_subalignment_without_allgap_columns method.
        It enables reading back an alignment that was previously written to
        disk, and relabeling its columns with the original labels.
        extension_path is a path to an .sto file (or a list of paths).
        Alignments from these .sto files are also read, and merged with the
        original (base) alignment.
        '''
        remaining_cols = self.annotations["ref.alignment.columns"]
        assert remaining_cols is not None and \
            len(remaining_cols) != 0, \
            ("Subproblem needs to have a proper list of alignment columns "
             "associated with it")

        _LOG.debug(
            "Reading %s %s and relabeling it based on %d orig column labels." %
            (orig_path, extension_path, len(remaining_cols)))

        ap_alg = ExtendedAlignment(list(self.fragments.keys()))
        ap_alg.build_extended_alignment(
            orig_path, extension_path, convert_to_string)
        ap_alg.relabel_original_columns(remaining_cols)
        return ap_alg



import os
import tempfile
import re
import sepp
from sepp.alignment import MutableAlignment
from sepp.alignment import _write_fasta
from sepp.config import options
'''
Collection of functions for metagenomic pipeline for taxonomic classification
Created on June 3, 2014

@author: namphuon
'''
global character_map, taxon_map, level_map, key_map, marker_genes, cog_genes
character_map = {'A': 'T', 'a': 't', 'C': 'G', 'c': 'g', 'T': 'A',
                 't': 'a', 'G': 'C', 'g': 'c', '-': '-'}
global levels
levels = ["species", "genus", "family", "order", "class", "phylum"]
marker_genes = [
    "nusA", "rplB", "rplK", "rplS", "rpsE", "rpsS",  "pgk", "rplC", "rplL",
    "rplT", "rpsI", "smpB", "dnaG", "pyrg",  "rplD", "rplM", "rpmA", "rpsJ",
    "frr", "pyrG1", "rplE", "rplN",  "rpsB", "rpsK", "infC", "rplA", "rplF",
    "rplP", "rpsC", "rpsM"]
cog_genes = [
    "COG0049", "COG0088", "COG0094", "COG0100", "COG0184", "COG0201",
    "COG0522", "COG0012", "COG0052", "COG0090", "COG0096", "COG0102",
    "COG0185", "COG0202", "COG0525", "COG0016", "COG0080", "COG0091",
    "COG0097", "COG0103", "COG0186", "COG0215", "COG0533", "COG0018",
    "COG0081", "COG0092", "COG0098", "COG0124", "COG0197", "COG0256",
    "COG0541", "COG0048", "COG0087", "COG0093", "COG0099", "COG0172",
    "COG0200", "COG0495", "COG0552"]


# TODO Fix parameter passing
# TODO Make taxonomy loading a class
def load_taxonomy(taxonomy_file, lower=True):
    global taxon_map, level_map, key_map
    f = open(taxonomy_file, 'r')

    # First line is the keywords for the taxonomy, need to map the keyword to
    # the positional index of each keyword
    results = f.readline().lower().replace('"', '').strip().split(',')
    key_map = dict([(results[i], i) for i in range(0, len(results))])

    # Now fill up taxonomy, level maps keep track of what taxa exist at each
    # level, taxon_map keep track of entire taxonomy
    taxon_map = {}
    level_map = {"species": {}, "genus": {}, "family": {}, "order": {},
                 "class": {}, "phylum": {}}

    for line in f:
        results = line.replace('"', '').strip()
        if (lower):
            results.lower()
        results = results.split(',')
        # insert into taxon map
        taxon_map[results[0]] = results

        # insert into level map
        for level in levels:
            if (results[key_map[level]] == ''):
                continue
            else:
                if (results[key_map[level]] not in level_map[level]):
                    level_map[level][results[key_map[level]]] = {}
                level_map[level][results[key_map[level]]][results[0]] = \
                    results[0]
    return (taxon_map, level_map, key_map)


def build_profile(input, output_directory):
    global taxon_map, level_map, key_map, levels
    temp_dir = tempfile.mkdtemp(dir=options().__getattribute__('tempdir'))
    if (options().bin == 'blast'):
        binned_fragments = blast_to_markers(input, temp_dir)
    else:
        binned_fragments = hmmer_to_markers(input, temp_dir)

    if binned_fragments:
        print("Finished binning")
    else:
        print("Unable to bin any fragments!\n")
        return

    # load up taxonomy for 30 marker genes
    if (options().genes == 'markers'):
        (taxon_map, level_map, key_map) = load_taxonomy(os.path.join(
            options().reference.path, 'refpkg/rpsB.refpkg/all_taxon.taxonomy'))
    else:
        (taxon_map, level_map, key_map) = load_taxonomy(os.path.join(
            options().reference.path,
            'refpkg/COG0012.refpkg/all_taxon.taxonomy'))

    # all classifications stored here
    classifications = {}
    classification_files = []
    # Now run TIPP on each fragment
    gene_name = 'sate'
    if (options().genes == 'cogs'):
        gene_name = 'pasta'
    for (gene, frags) in binned_fragments.items():
        # Get size of each marker
        total_taxa = 0
        with open(os.path.join(options().__getattribute__('reference').path,
                  'refpkg/%s.refpkg/%s.size' % (gene, gene_name)), 'r') as f:
            total_taxa = int(f.readline().strip())
        decomp_size = options().alignment_size
        if (decomp_size > total_taxa):
            decomp_size = int(total_taxa / 2)
        cpus = options().cpu
        if (len(frags) < cpus):
            cpus = len(frags)
        extra = ''
        if options().dist is True:
            extra = '-D'
        if options().max_chunk_size is not None:
            extra = extra + '-F %d' % options().max_chunk_size
        if options().cutoff != 0:
            extra = extra+" -C %f" % options().cutoff
        print(
            ('Cmd:\nrun_tipp.py -c %s --cpu %s -m %s -f %s -t %s -adt %s -a '
             '%s -r %s -tx %s -txm %s -at %0.2f -pt %0.2f -A %d -P %d -p %s '
             '-o %s -d %s %s') %
            (options().config_file.name,
             cpus,
             options().molecule,
             temp_dir+"/%s.frags.fas.fixed" % gene,
             os.path.join(options().__getattribute__('reference').path,
                          'refpkg/%s.refpkg/%s.taxonomy' % (gene, gene_name)),
             os.path.join(options().__getattribute__('reference').path,
                          'refpkg/%s.refpkg/%s.tree' % (gene, gene_name)),
             os.path.join(options().__getattribute__('reference').path,
                          'refpkg/%s.refpkg/%s.fasta' % (gene, gene_name)),
             os.path.join(options().__getattribute__('reference').path,
                          'refpkg/%s.refpkg/%s.taxonomy.RAxML_info' % (
                            gene, gene_name)),
             os.path.join(options().__getattribute__('reference').path,
                          'refpkg/%s.refpkg/all_taxon.taxonomy' % gene),
             os.path.join(options().__getattribute__('reference').path,
                          'refpkg/%s.refpkg/species.mapping' % gene),
             options().alignment_threshold,
             0,
             decomp_size,
             total_taxa,
             temp_dir+"/temp_file",
             "tipp_%s" % gene,
             output_directory+"/markers/",
             extra))

        os.system(
            ('run_tipp.py -c %s --cpu %s -m %s -f %s -t %s -adt %s -a %s -r %s'
             ' -tx %s -txm %s -at %0.2f -pt %0.2f -A %d -P %d -p %s -o %s -d '
             '%s %s') %
            (options().config_file.name,
             cpus,
             options().molecule,
             temp_dir+"/%s.frags.fas.fixed" % gene,
             os.path.join(options().__getattribute__('reference').path,
                          'refpkg/%s.refpkg/%s.taxonomy' % (gene, gene_name)),
             os.path.join(options().__getattribute__('reference').path,
                          'refpkg/%s.refpkg/%s.tree' % (gene, gene_name)),
             os.path.join(options().__getattribute__('reference').path,
                          'refpkg/%s.refpkg/%s.fasta' % (gene, gene_name)),
             os.path.join(options().__getattribute__('reference').path,
                          'refpkg/%s.refpkg/%s.taxonomy.RAxML_info' % (
                            gene, gene_name)),
             os.path.join(options().__getattribute__('reference').path,
                          'refpkg/%s.refpkg/all_taxon.taxonomy' % gene),
             os.path.join(options().__getattribute__('reference').path,
                          'refpkg/%s.refpkg/species.mapping' % gene),
             options().alignment_threshold,
             0,
             decomp_size,
             total_taxa,
             temp_dir+"/temp_file",
             "tipp_%s" % gene,
             output_directory+"/markers/",
             extra))
        if (not os.path.exists(output_directory +
                               "/markers/tipp_%s_classification.txt" % gene)):
            continue

        gene_classification = generate_classification(
            output_directory + "/markers/tipp_%s_classification.txt" % gene,
            options().placement_threshold)
        classification_files.append(
            output_directory + "/markers/tipp_%s_classification.txt" % gene)
        # Now write individual classification and also pool classifications
        write_classification(
            gene_classification,
            output_directory + "/markers/tipp_%s.classification" % gene)
        classifications.update(gene_classification)
    remove_unclassified_level(classifications)
    write_classification(classifications,
                         output_directory+"/markers/all.classification")
    write_abundance(classifications, output_directory)

    if (options().dist is True):
        distribution(classification_files, output_directory)


def distribution(classification_files, output_dir):
    global taxon_map, level_map, key_map, levels, level_names
    distribution = {"species": {}, "genus": {}, "family": {}, "order": {},
                    "class": {}, "phylum": {}}
    total_frags = 0
    for class_input in classification_files:
        class_in = open(class_input, 'r')
        frag_info = {"species": {'unclassified': 1},
                     "genus": {'unclassified': 1},
                     "family": {'unclassified': 1},
                     "order": {'unclassified': 1},
                     "class": {'unclassified': 1},
                     "phylum": {'unclassified': 1}}
        old_name = ""
        for line in class_in:
            results = line.strip().split(',')
            if (len(results) > 5):
                results = [results[0], results[1], results[2],
                           results[-2], results[-1]]
            (name, id, rank, probability) = (
                results[0], results[1], results[3], float(results[4]))
            if (rank not in distribution):
                continue
            if (old_name == ""):
                old_name = name
            if (name != old_name):
                total_frags += 1
                assert frag_info['phylum']['unclassified'] != 1
                for clade, cladeval in frag_info.items():
                    for clade_name, cnc in cladeval.items():
                        if (clade_name, cnc not in distribution[clade]):
                            distribution[clade][clade_name] = 0
                        distribution[clade][clade_name] += cnc
                frag_info = {"species": {'unclassified': 1},
                             "genus": {'unclassified': 1},
                             "family": {'unclassified': 1},
                             "order": {'unclassified': 1},
                             "class": {'unclassified': 1},
                             "phylum": {'unclassified': 1}}
                old_name = name
            if (id not in frag_info[rank]):
                frag_info[rank][id] = 0
            frag_info[rank][id] += probability
            frag_info[rank]['unclassified'] -= probability
        total_frags += 1
        assert frag_info['phylum']['unclassified'] != 1
        for clade, cladeval in frag_info.items():
            for clade_name, cnc in cladeval.items():
                if (clade_name not in distribution[clade]):
                    distribution[clade][clade_name] = 0
                distribution[clade][clade_name] += cnc

    level_names = {1: 'species', 2: 'genus', 3: 'family', 4: 'order',
                   5: 'class', 6: 'phylum'}
    for level in level_names:
        f = open(output_dir + "/abundance.distribution.%s.csv" %
                 level_names[level], 'w')
        f.write('taxa\tabundance\n')
        lines = []
        for clade, value in distribution[level_names[level]].items():
            name = clade
            if (name != 'unclassified'):
                name = taxon_map[clade][key_map['tax_name']]
            lines.append('%s\t%0.4f\n' % (name, float(value) / total_frags))
        lines.sort()
        f.write(''.join(lines))
        f.close()
    return distribution


def remove_unclassified_level(classifications, level=6):
    global taxon_map, level_map, key_map, levels
    frags = list(classifications.keys())
    for frag in frags:
        if classifications[frag][level] == 'NA':
            del classifications[frag]


def write_classification(class_input, output):
    '''Writes a classification file
    '''
    class_out = open(output, 'w')
    class_out.write("fragment\tspecies\tgenus\tfamily\torder\tclass\tphylum\n")
    keys = list(class_input.keys())
    keys.sort()
    for frag in keys:
        class_out.write("%s\n" % "\t".join(class_input[frag]))
    class_out.close()


# Fix problem with NA being unclassified
def write_abundance(classifications, output_dir, labels=True,
                    remove_unclassified=True):
    global taxon_map, level_map, key_map, levels

    level_abundance = {
        1: {'total': 0}, 2: {'total': 0}, 3: {'total': 0}, 4: {'total': 0},
        5: {'total': 0}, 6: {'total': 0}}

    level_names = {1: 'species', 2: 'genus', 3: 'family', 4: 'order',
                   5: 'class', 6: 'phylum'}
    for lineage in classifications.values():
        # insert into level map
        for level in range(1, 7):
            if (lineage[level] == 'NA'):
                if ('unclassified' not in level_abundance[level]):
                    level_abundance[level]['unclassified'] = 0
                level_abundance[level]['unclassified'] += 1
                level_abundance[level]['total'] += 1
                # continue
            else:
                if (lineage[level] not in level_abundance[level]):
                    level_abundance[level][lineage[level]] = 0
                level_abundance[level][lineage[level]] += 1
                level_abundance[level]['total'] += 1
    for level in level_names:
        f = open(output_dir + "/abundance.%s.csv" % level_names[level], 'w')
        f.write('taxa\tabundance\n')
        lines = []
        for clade in level_abundance[level]:
            if clade == 'total':
                continue
            name = clade
            if labels and name != 'unclassified':
                name = taxon_map[clade][key_map['tax_name']]
            lines.append('%s\t%0.4f\n' % (
                name, float(level_abundance[level][clade]) / level_abundance[
                    level]['total']))
        lines.sort()
        f.write(''.join(lines))
        f.close()


def generate_classification(class_input, threshold):
    global taxon_map, level_map, key_map, levels
    class_in = open(class_input, 'r')
    level_map_hierarchy = {"species": 0, "genus": 1, "family": 2, "order": 3,
                           "class": 4, "phylum": 5, "root": 6}
    # Need to keep track of last line so we can determine when we switch to
    # new classification
    old_name = ""
    old_probability = 1
    old_id = ""
    old_rank = ""

    # keep track of all fragment names
    names = {}
    classification = {}
    for line in class_in:
        results = line.strip().split(',')
        if (len(results) > 5):
            results = [results[0], results[1], results[2],
                       results[-2], results[-1]]
        (name, id, rank, probability) = (
            results[0], results[1], results[3], float(results[4]))
        names[name] = name
        if (name != old_name):
            # when we switch to new fragment, output last classification for
            # old fragment
            if (old_name != ""):
                lineage = taxon_map[old_id]
                output_line = [old_name]
                for level in levels:
                    clade = lineage[key_map[level]]
                    if (clade == ""):
                        clade = "NA"
                    output_line.append(clade)
                classification[old_name] = output_line
            old_name = name
            old_rank = "root"
            old_probability = 1
            old_id = '1'

        # Switch to new rank if the new probability is higher than threshold
        # and our rank is more specific than our original rank
        if (
            rank in level_map_hierarchy and
            (level_map_hierarchy[old_rank] > level_map_hierarchy[rank]) and
            (probability > threshold)
           ):
            old_rank = rank
            old_probability = probability
            old_id = id
        # Switch to new rank if the new rank matches old rank but has higher
        # probability
        elif (
              rank in level_map_hierarchy and
              (level_map_hierarchy[old_rank] == level_map_hierarchy[rank]) and
              (probability > old_probability)
             ):
            old_rank = rank
            old_probability = probability
            old_id = id

    if old_id in taxon_map:
        lineage = taxon_map[old_id]
        output_line = [old_name]
        for level in levels:
            clade = lineage[key_map[level]]
            if (clade == ""):
                clade = "NA"
            output_line.append(clade)
        classification[name] = output_line
    return classification


def hmmer_to_markers(input, temp_dir):
    global marker_genes
    fragments = MutableAlignment()
    fragments.read_filepath(input)

    reverse = dict([(name+'_rev', reverse_sequence(seq))
                    for (name, seq) in fragments.items()])
    all_frags = MutableAlignment()
    all_frags.set_alignment(fragments)
    all_frags.set_alignment(reverse)
    frag_file = temp_dir+"/frags.fas"
    _write_fasta(all_frags, frag_file)

    # Now bin the fragments
    frag_scores = dict([(name, [-10000, 'NA', 'NA'])
                        for name in fragments.keys()])
    gene_set = marker_genes
    align_name = 'sate'
    if (options().genes == 'cogs'):
        gene_set = cog_genes
        align_name = 'pasta'
    for gene in gene_set:
        # Now run HMMER search
        hmmer_search(
            frag_file,
            os.path.join(
                options().__getattribute__('reference').path,
                'refpkg/%s.refpkg/%.profile' % (gene, align_name)),
            temp_dir + "/%s.out" % gene)
        results = read_hmmsearch_results(temp_dir + "/%s.out" % gene)

        # Now select best direction for each frag
        for name, value in results.items():
            bitscore = value[1]
            direction = 'forward'
            true_name = name
            if (name.find('_rev') != -1):
                true_name = true_name.replace('_rev', '')
                direction = 'reverse'
            if frag_scores[true_name][0] < bitscore:
                frag_scores[true_name] = [bitscore, gene, direction]

    # Now bin the fragments
    genes = dict([])
    for name, val in frag_scores.items():
        if (val[1] not in genes):
            genes[val[1]] = {}
        if (val[2] == 'forward'):
            genes[val[1]][name] = fragments[name]
        else:
            genes[val[1]][name] = reverse_sequence(fragments[name])
    genes.pop("NA", None)
    for gene, seq in genes.items():
        gene_file = temp_dir + "/%s.frags.fas" % gene
        _write_fasta(seq, gene_file + ".fixed")
    return genes


def blast_to_markers(input, temp_dir):
    fragments = MutableAlignment()
    fragments.read_filepath(input)

    if (options().gene is None):
        # First blast sequences against all markers
        blast_results = temp_dir + "/blast.out"
        if (options().blast_file is None):
            print("Blasting fragments against marker dataset\n")
            blast_fragments(input, blast_results)
        else:
            blast_results = options().blast_file
        # Next bin the blast hits to the best gene
        gene_binning = bin_blast_results(blast_results)
    else:
        gene_binning = {options().gene: list(fragments.keys())}
    # Now figure out direction of fragments
    binned_fragments = dict([
        (gene, dict([(seq_name, fragments[seq_name])
                     for seq_name in gene_binning[gene]]))
        for gene in gene_binning])
    print("Finding best orientation of reads\n")
    align_name = 'sate'
    if (options().genes == 'cogs'):
        align_name = 'pasta'
    for (gene, frags) in binned_fragments.items():
        # Add reverse complement sequence
        frags_rev = dict([(name + '_rev', reverse_sequence(seq))
                          for (name, seq) in frags.items()])
        gene_frags = MutableAlignment()
        gene_frags.set_alignment(frags)
        gene_frags.set_alignment(frags_rev)
        gene_file = temp_dir + "/%s.frags.fas" % gene
        _write_fasta(gene_frags, gene_file)

        # Now run HMMER search
        hmmer_search(
            gene_file,
            os.path.join(
                options().__getattribute__('reference').path,
                'refpkg/%s.refpkg/%s.hmm' % (gene, align_name)),
            temp_dir + "/%s.out" % gene)
        results = read_hmmsearch_results(temp_dir + "/%s.out" % gene)

        # Now select best direction for each frag
        for key in frags:
            forward_score = -10000
            backward_score = -10000
            if (key in results):
                forward_score = results[key][1]
            if (key+"_rev" in results):
                backward_score = results[key + "_rev"][1]
            if (backward_score > forward_score):
                frags[key] = gene_frags[key + "_rev"]

        # Now write to file
        _write_fasta(frags, gene_file + ".fixed")
        binned_fragments[gene] = frags
    return binned_fragments


def read_hmmsearch_results(input):
    # Group 1 (e-value) 2 (bitscore) and 9 (taxon name) contain the
    # relevant information, other ones can be ignored unless we plan to do
    # something later
    pattern = re.compile(
        r"([^\s]+)\s+([^\s]+)\s+([^\s]+)\s+([^\s]+)\s+([^\s]+)\s+([^\s]+)"
        r"\s+([^\s]+)\s+([^\s]+)\s+([^\s]+)")
    start_reading = False
    infile = open(input)
    results = {}
    for line in infile:
        line = line.strip()
        if (not start_reading and line.startswith("E-value") is True):
            start_reading = True
        elif (start_reading and line == ""):
            start_reading = False
            break
        elif (start_reading):
            matches = pattern.search(line)
            if (matches is not None and matches.group(0).find("--") == -1):
                results[matches.group(9).strip()] = (
                    float(matches.group(1).strip()),
                    float(matches.group(2).strip()))
    return results


def read_mapping(input, header=False, delimiter='\t'):
    '''Read a mapping file
    '''
    d = {}
    with open(input) as f:
        for line in f:
            if (header is True):
                next
            results = line.strip().split(delimiter)
            d[results[0]] = results
    return d


def bin_blast_results(input):
    # Map the blast results to the markers
    gene_mapping = read_mapping(
        os.path.join(
            options().__getattribute__('reference').path,
            'blast/%s/seq2marker.tab' % options().genes))

    genes = {}
    with open(input) as f:
        for line in f:
            results = line.split('\t')
            gene = gene_mapping[results[1]][1]
            if gene in genes:
                genes[gene].append(results[0])
            else:
                genes[gene] = [results[0]]
    return genes


def hmmer_search(input, hmmer, output):
    '''Blast the fragments against all marker genes+16S sequences, return
    output'''
    os.system('%s --noali -E 10000 --cpu %d -o %s %s %s' % (
        options().__getattribute__('hmmsearch').path,
        options().cpu, output, hmmer, input))


def blast_fragments(input, output):
    '''Blast the fragments against all marker genes+16S sequences, return
    output'''
    os.system(
        ('%s -db %s -outfmt 6 -query %s -out %s -num_threads %d '
         '-max_target_seqs 1 ') %
        (options().__getattribute__('blast').path,
         os.path.join(
            options().__getattribute__('reference').path,
            "blast/%s/alignment.fasta.db" % options().genes),
         input, output, options().cpu))


def reverse_sequence(sequence):
    global character_map
    #  Reverse complement the sequence
    return "".join([character_map.get(a, a) for a in sequence[::-1]])


def augment_parser():
    # default_settings['DEF_P'] = (100 , "Number of taxa (i.e.
    # no decomposition)")
    parser = sepp.config.get_parser()

    tippGroup = parser.add_argument_group(
        "TIPP Options".upper(),
        "These arguments set settings specific to TIPP")

    tippGroup.add_argument(
        "-at", "--alignmentThreshold", type=float,
        dest="alignment_threshold", metavar="N",
        default=0.0,
        help="Enough alignment subsets are selected to reach a commulative "
             "probability of N. "
             "This should be a number between 0 and 1 [default: 0.95]")

    tippGroup.add_argument(
        "-pt", "--placementThreshold", type=float,
        dest="placement_threshold", metavar="N",
        default=0.0,
        help="Enough placements are selected to reach a commulative "
             "probability of N. "
             "This should be a number between 0 and 1 [default: 0.95]")
    tippGroup.add_argument(
        "-g", "--gene", type=str,
        dest="gene", metavar="N",
        default=None,
        help="Classify on only the specified gene. ")

    tippGroup.add_argument(
        "-b", "--blast_file", type=str,
        dest="blast_file", metavar="N",
        default=None,
        help="Blast file with fragments already binned. ")

    tippGroup.add_argument(
        "-bin", "--bin_using", type=str,
        dest="bin", metavar="N",
        default="blast",
        help="Tool for binning")

    tippGroup.add_argument(
        "-D", "--dist",
        dest="dist", action='store_true',
        default=False,
        help="Treat fragments as distribution")

    tippGroup.add_argument(
        "-C", "--cutoff", type=float,
        dest="cutoff", metavar="N",
        default=0.0,
        help="Placement probability requirement to count toward the "
             "distribution. "
             "This should be a number between 0 and 1 [default: 0.0]")

    tippGroup.add_argument(
        "-G", "--genes", type=str,
        dest="genes", metavar="GENES",
        default='markers',
        help="Use markers or cogs genes [default: markers]")


def main():
    augment_parser()
    sepp.config._options_singelton = sepp.config._parse_options()
    if (options().alignment_size is None):
        options().alignment_size = 100
    input = options().fragment_file.name
    output_directory = options().outdir
    build_profile(input, output_directory)


if __name__ == '__main__':
    main()



'''
Created on Sep 19, 2012

@author: smirarab
'''
import unittest
from sepp.alignment import MutableAlignment, ReadonlySubalignment,\
    ExtendedAlignment
from sepp.problem import SeppProblem
from sepp.filemgr import get_data_path
from tempfile import mkstemp
from os import remove


class Test(unittest.TestCase):
    fp_dummy1 = None
    fp_dummy2 = None
    fp_dummy3 = None
    fp_dummy4 = None
    fp_dummy5 = None

    def setUp(self):
        _, self.fp_dummy1 = mkstemp()
        _, self.fp_dummy2 = mkstemp()
        _, self.fp_dummy3 = mkstemp()
        _, self.fp_dummy4 = mkstemp()
        _, self.fp_dummy5 = mkstemp()

    def tearDown(self):
        remove(self.fp_dummy1)
        remove(self.fp_dummy2)
        remove(self.fp_dummy3)
        remove(self.fp_dummy4)
        remove(self.fp_dummy5)

    def testAlignmentReadFasta(self):
        alg = MutableAlignment()
        alg.read_filepath(get_data_path("mock/pyrg/sate.fasta"))

        assert len(alg) == 65, "MutableAlignment length is %s" % len(alg)

        assert all([not alg.is_all_gap(i) for i in range(0, alg.get_length())])

    def testReadOnlySubAlignment(self):
        alg = MutableAlignment()
        alg.read_filepath(get_data_path("mock/pyrg/sate.fasta"))

        subset = ['NC_008701_720717_722309', 'NC_013156_149033_150643',
                  'NC_013887_802739_801129']
        readonly_subalignment = ReadonlySubalignment(subset, alg)

        assert len(readonly_subalignment) == 3, len(readonly_subalignment)

        assert set(readonly_subalignment.keys()) == set(
            readonly_subalignment.get_sequence_names()) == set(subset), \
            "Subalignment keys not matching given keys %s vs %s" % (
            list(readonly_subalignment.keys()), subset)

        for (k, s) in list(readonly_subalignment.items()):
            assert k in subset, \
                "%s not found in subset but returned by subalignment" % k
            assert s == alg[k], \
                "sequence associated with %k not matching parent alignment" % k

        try:
            readonly_subalignment[2] = "ACGT"
            assert False, "Readony alignment is successfully modified. "
        except TypeError:
            pass

        assert readonly_subalignment.get_length() == alg.get_length(), \
            "alignment length should not change"

        assert readonly_subalignment.is_aligned() is True

        assert readonly_subalignment.is_all_gap(2) is True, \
            "Site 2 should be all gaps"
        assert readonly_subalignment.is_all_gap(150) is False, \
            "Site 100 should not be all gaps"

        readonly_subalignment.write_to_path(
            self.fp_dummy1)  # "mock/pyrg/sate.sub.fasta"

        mutable_subalignment = readonly_subalignment.get_mutable_alignment()
        mutable_subalignment.delete_all_gap()

        assert all([not mutable_subalignment.is_all_gap(i)
                    for i in range(0, mutable_subalignment.get_length())])

    def testExtendedAlignment(self):
        subset = [
            "SFIF", "SFII", "SCFC", "SGHD", "SDCC", "SBGE", "SFBB", "SDI",
            "SCGB", "SJGF", "SGBI", "SCJA", "SGAD", "SHEB", "SFHB", "SDJI",
            "SHED", "SJJJ", "SBBE", "SCCH", "SDJB", "SDAC", "SHEH", "SFDC",
            "SFEI", "SHHB", "SC", "SIAB", "SDDI", "SBCB", "SJB", "SEBD",
            "SFGD", "SHA", "SIDA", "SGHI", "SGIB", "SBFJ", "SFIE", "SCJF",
            "SJHJ", "SJBG", "SEJI", "SFFF", "SJ", "SIII", "SJHH", "SEIH",
            "SBDC", "SHDJ", "SJDD", "SGDB", "SIHA", "SIBB", "SECC", "SCAD",
            "SGBB", "SGIF", "SJHC", "SFCD", "SEAA", "SEFF", "SDFG", "SDJE",
            "SCFG", "SFH", "SCJ", "SDDD", "SEGD", "SCIH", "SDAG", "SCJE",
            "SFAJ", "SIDJ", "SE", "SHBC", "SJFF", "SCHD", "SBHA", "SEDF",
            "SFAF", "SEDD", "SDHD", "SGJD", "SIBH", "SGDF", "SIFA", "SJGA",
            "SIJB", "SFI", "SGA", "SBFC", "SBJA", "SFFC", "SFDH", "SFEE",
            "SBDF", "SGBJ", "SDHE", "SJIB", "SHHI", "SIDE", "SJII"]

        alg = MutableAlignment()
        alg.read_filepath(get_data_path("simulated/test.fasta"))
        alg.delete_all_gap()
        tlen = alg.get_length()

        frg = MutableAlignment()
        frg.read_filepath(get_data_path("simulated/test.fas"))
        # print frg.get_num_taxa()

        pp = SeppProblem(list(alg.keys()))
        pp.fragments = frg
        pp.subalignment = alg

        cp1 = SeppProblem(subset, pp)
        cp2 = SeppProblem(list(set(alg.keys()) - set(subset)), pp)
        cp1.fragments = ReadonlySubalignment(
            [k for k in list(frg.keys()) if int(k[-1]) >= 9], frg)
        cp2.fragments = ReadonlySubalignment(
            [k for k in list(frg.keys()) if int(k[-1]) <= 1], frg)

        cp1labels = cp1.write_subalignment_without_allgap_columns(
            self.fp_dummy1)  # tmp/cp1.fasta
        cp2labels = cp2.write_subalignment_without_allgap_columns(
            self.fp_dummy2)  # tmp/cp2.fasta
        tmp = MutableAlignment().read_filepath(self.fp_dummy1)
        assert all([not tmp.is_all_gap(pos)
                    for pos in range(0, tmp.get_length())])
        tmp = MutableAlignment().read_filepath(self.fp_dummy2)
        assert all([not tmp.is_all_gap(pos)
                    for pos in range(0, tmp.get_length())])

        cp1.fragments.write_to_path(self.fp_dummy3)  # tmp/cp1.frags.fas
        cp2.fragments.write_to_path(self.fp_dummy4)  # tmp/cp2.frags.fas

        '''We have done the hmmalign before.
           don't worry about that right now'''

        ext1 = ExtendedAlignment(cp1.fragments)
        ext1.build_extended_alignment(
            self.fp_dummy1,
            get_data_path("tmp/cp1.extended.sto"))
        ext1.relabel_original_columns(cp1labels)
        ext2 = ExtendedAlignment(cp2.fragments)
        ext2.build_extended_alignment(
            self.fp_dummy2,
            get_data_path("tmp/cp2.extended.sto"))
        ext2.relabel_original_columns(cp2labels)

        extmerger = ExtendedAlignment([])
        extmerger.merge_in(ext1)
        mixed = extmerger.merge_in(ext2)

        extmerger.write_to_path(self.fp_dummy5)  # tmp/extended.merged.fasta

        assert extmerger.is_aligned(), "Merged alignment is not aligned"
        in1 = len([x for x in ext1._col_labels if x < 0])
        in2 = len([x for x in ext2._col_labels if x < 0])
        assert (in1 + in2 + tlen - mixed) == extmerger.get_length(), \
            ("Lengths don't match up after merging. Merged:%d. Insertion1:%d "
             "Insertion2:%d BaseLen:%d Mixed-insertion: %d") % (
                extmerger.get_length(), in1, in2, tlen, mixed)
        assert (in1 + in2 - mixed) == len(extmerger.get_insertion_columns()), \
            ("Columns are not correctly labeled after merging. Merged "
             "insertion count:%d. Insertion1:%d Insertion2:%d Mixed-insertion:"
             " %d") % (
             len(list(extmerger.iter_insertion_columns())), in1, in1, mixed)

        tmp = extmerger.get_base_readonly_alignment().get_mutable_alignment()
        tmp.delete_all_gap()
        assert tmp.is_aligned(), "merged alignment should be aligned!"
        assert tmp.get_length() == tlen, "merged alignment has wrong length"
        assert all([alg[k] == s for (k, s) in list(tmp.items())]), \
            "merged alignment should match original alignment"


if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testAlignmentReadFasta']
    unittest.main()



from sepp.exhaustive import ExhaustiveAlgorithm, JoinAlignJobs
from sepp.exhaustive import get_placement_job_name
import sepp
import os
import stat
import math
from sepp.config import options
import argparse
import json
from sepp.algorithm import AbstractAlgorithm
from sepp.alignment import MutableAlignment, ExtendedAlignment
from sepp.jobs import HMMBuildJob, HMMSearchJob, HMMAlignJob, PplacerJob,\
    ExternalSeppJob
from sepp.problem import SeppProblem
from sepp.scheduler import JobPool, Join
from sepp.tree import PhylogeneticTree
from dendropy.datamodel.treemodel import Tree
import dendropy
import pickle
from sepp import get_logger
from functools import reduce

_LOG = get_logger(__name__)


class TIPPJoinSearchJobs(Join):
    '''
    After all search jobs have finished on tips, we need to figure out which
    fragment goes to which subset and start aligning fragments.
    This join takes care of that step.
    '''
    def __init__(self, alignment_threshold):
        Join.__init__(self)
        self.alignment_threshold = alignment_threshold

    def setup_with_root_problem(self, root_problem):
        self.root_problem = root_problem
        for p in root_problem.iter_leaves():
            self.add_job(p.jobs["hmmsearch"])

    def figureout_fragment_subset(self):
        ''' Figure out which fragment should go to which subproblem'''
        # We need to keep and check the following flag because of
        # checkpoining scenarios (join already done before!)
        if "fragments.distribution.done" in self.root_problem.annotations:
            return
        bitscores = dict([(name, []) for name in list(
            self.root_problem.fragments.keys())])
        for fragment_chunk_problem in self.root_problem.iter_leaves():
            align_problem = fragment_chunk_problem.get_parent()
            assert isinstance(align_problem, SeppProblem)
            '''For each subproblem start with an empty set of fragments,
            and add to them as we encounter new best hits for that
            subproblem'''
            if align_problem.fragments is None:
                align_problem.fragments = MutableAlignment()
            search_res = fragment_chunk_problem.get_job_result_by_name(
                "hmmsearch")
            for key in list(search_res.keys()):
                ''' keep a list of all hits, and their bit scores'''
                bitscores[key].append((search_res[key][1], align_problem))

        for frag, tuplelist in bitscores.items():
            ''' TODO: what to do with those that are not? For now, only output
                warning message'''
            # TODO:  Need to double check and fix the math
            _LOG.warning("Fragment %s is not scored against any subset" %
                         str(frag))
            if len(tuplelist) == 0:
                _LOG.warning("Fragment %s is not scored against any subset" %
                             str(frag))
                continue
            ''' convert bit scores to probabilities '''
            denum = sum(math.pow(2, min(x[0], 1022)) for x in tuplelist)
            # _LOG.warning("Tuples: %s" %str(tuplelist))
            tuplelist = [
                ((math.pow(2, min(x[0], 1022)) / denum * 1000000), x[1])
                for x in tuplelist]
            ''' Sort subsets by their probability'''
            tuplelist.sort(reverse=True, key=lambda x: x[0])
            ''' Find enough subsets to reach the threshold '''
            selected = tuplelist[
                0: max(1, reduce(
                    lambda x, y: (x[0], None)
                    if x[1] is None else
                    (y[0], x[1] + y[1]) if x[1] < int(
                        1000000 * self.alignment_threshold) else
                    (y[0], None),
                    enumerate([x[0] for x in tuplelist]))[0])]

            ''' Renormalize the selected list to add up to 1'''
            renorm = 0
            for (prob, align_problem) in selected:
                renorm = renorm + prob / 1000000
            renorm = 1 / renorm

            _LOG.debug("Fragment %s assigned to %d subsets" %
                       (frag, len(selected)))
            ''' Rename the fragment and assign it to the respective subsets'''
            for (prob, align_problem) in selected:
                postfix = prob * renorm if \
                    options().exhaustive.weight_placement_by_alignment.\
                    lower() == "true" \
                    else 1000000
                frag_rename = "%s_%s_%d" % (frag, align_problem.label, postfix)
                align_problem.fragments[frag_rename] = \
                    self.root_problem.fragments[frag]

        self.root_problem.annotations["fragments.distribution.done"] = 1

    def perform(self):
        '''
        Distributes fragments to alignments subsets with best score,
        and runs align jobs on those. Also, creates new chunks of fragments
        for better parallelism.
        '''

        ''' Figure out which fragment should go to which subproblem'''
        self.figureout_fragment_subset()

        ''' For each alignment subproblem,
        1) make sure its fragments are evenly distributed to fragment chunks.
        2) Setup alignment jobs for its children and enqueue them'''
        alg_problems = [alg
                        for p in self.root_problem.children
                        for alg in p.children]
        for alg_problem in alg_problems:
            assert isinstance(alg_problem, SeppProblem)
            chunks = len(alg_problem.get_children())
            fragment_chunks = alg_problem.fragments.divide_to_equal_chunks(
                chunks)

            ''' Now setup alignment jobs and enqueue them'''
            for (i, fragment_chunk_problem) in enumerate(alg_problem.children):
                fragment_chunk_problem.fragments = fragment_chunks[i]
                aj = fragment_chunk_problem.jobs['hmmalign']
                assert isinstance(aj, HMMAlignJob)
                ''' First Complete setting up alignments'''
                aj.hmmmodel = alg_problem.get_job_result_by_name('hmmbuild')
                aj.base_alignment = alg_problem.jobs["hmmbuild"].infile

                if fragment_chunk_problem.fragments is None \
                   or fragment_chunk_problem.fragments.is_empty():
                    aj.fake_run = True
                else:
                    fragment_chunk_problem.fragments.write_to_path(
                        aj.fragments)
                ''' Now the align job can be put on the queue '''
                JobPool().enqueue_job(aj)

    def __str__(self):
        return "join search jobs for all tips of ", self.root_problem


class TIPPJoinAlignJobs(JoinAlignJobs):
    '''
    After all alignments jobs for a placement subset have finished,
    we need to build those extended alignments and start placing fragments.
    This join takes care of that step.
    '''
    def __init__(self, placer):
        Join.__init__(self)
        self.placer = placer

    def perform(self):
        pp = self.placement_problem
        fullExtendedAlignments = self.merge_subalignments()

        for i in range(0, self.root_problem.fragment_chunks):
            fullExtendedAlignment = fullExtendedAlignments[i]
            # Split the backbone alignment and query sequences into
            # separate files
            queryExtendedAlignment = \
                fullExtendedAlignment.get_fragments_readonly_alignment()
            baseAlignment = fullExtendedAlignment.get_base_readonly_alignment()
            pj = pp.jobs[get_placement_job_name(i)]

            if (queryExtendedAlignment.is_empty()):
                pj.fake_run = True

            if self.placer == "pplacer":
                assert isinstance(pj, PplacerJob)

                # Write out the extended alignments, split into query and
                # full-length for pplacer
                queryExtendedAlignment.write_to_path(
                    pj.extended_alignment_file)
                baseAlignment.write_to_path(pj.backbone_alignment_file)

            elif self.placer == "epa":
                # assert isinstance(pj, EPAJob)
                raise ValueError("EPA Currently not supported")

                # Write out the extended alignments in phylip for EPA
                fullExtendedAlignment.write_to_path(
                    pj.extended_alignment_file, schema="PHYLIP")

            # keep the extended alignment on everything
            # pj.set_attribute("full_extended_alignment_object",
            # fullExtendedAlignment)

            # TODO: Removed this, as it can cause unexpected lockups
            output = open(pj.full_extended_alignment_file, 'wb')
            pickle.dump(fullExtendedAlignment, output)
            output.close()

            # Enqueue the placement job
            JobPool().enqueue_job(pj)

    def __str__(self):
        return "join align jobs for tips of ", self.placement_problem


class TIPPMergeJsonJob(ExternalSeppJob):
    def __init__(self, **kwargs):
        self.job_type = 'jsonmerger'
        ExternalSeppJob.__init__(self, self.job_type, **kwargs)
        self.out_file = None
        self.distribution = False
        self.taxonomy = None
        self.mapping = None
        self.threshold = None
        self.classification_file = None
        self.elim = float(options().hmmsearch.elim)
        if options().hmmsearch.filters.upper() == "TRUE":
            self.filters = True
        else:
            if options().hmmsearch.filters.upper() == "FALSE":
                self.filters = False
            else:
                self.filters = None
        if self.filters is None:
            raise Exception(
                "Expecting true/false for options().hmmsearch.filters")
        self.strategy = options().exhaustive.strategy
        self.minsubsetsize = int(options().exhaustive.minsubsetsize)
        self.alignment_threshold = float(options().alignment_threshold)
        self.molecule = options().molecule
        self.placer = options().exhaustive.__dict__['placer'].lower()
        self.cutoff = 0

    def setup(self, inString, output_file, **kwargs):
        self.stdindata = inString
        self.out_file = output_file
        self._kwargs = kwargs

    def setup_for_tipp(self, inString, output_file, taxonomy, mapping,
                       threshold, classification_file, push_down,
                       distribution=False, cutoff=0, **kwargs):
        self.stdindata = inString
        self.out_file = output_file
        self.taxonomy = taxonomy.name
        self.mapping = mapping.name
        self.distribution = distribution
        self.threshold = str(threshold)
        self.classification_file = classification_file
        self.push_down = push_down
        self._kwargs = kwargs
        self.cutoff = cutoff

    def get_invocation(self):
        invoc = ["java", "-jar", self.path, "-", "-", self.out_file, "-r", "4"]
        if self.taxonomy is not None:
            invoc.extend(["-t", self.taxonomy])
        if self.mapping is not None:
            invoc.extend(["-m", self.mapping])
        if self.threshold is not None:
            invoc.extend(["-p", self.threshold])
        if self.classification_file is not None:
            invoc.extend(["-c", self.classification_file])
        if self.distribution:
            invoc.extend(["-d"])
        if not self.push_down:
            invoc.extend(["-u"])
        invoc.extend(["-C", str(self.cutoff)])
        # print " ".join(invoc)
        return invoc

    def characterize_input(self):
        return "input:pipe output:%s; Pipe:\n%s" % (
            self.out_file, self.stdindata)

    def read_results(self):
        '''
        Since the output file can be huge, we don't want to read it here,
        because it will need to get pickled and unpickled. Instead, we just
        send back the file name, and will let the caller figure out what to do
        with it.
        '''
        assert os.path.exists(self.out_file)
        assert os.stat(self.out_file)[stat.ST_SIZE] != 0
        return (self.out_file, self.stdoutdata)


class TIPPExhaustiveAlgorithm(ExhaustiveAlgorithm):
    '''
    This implements the exhaustive algorithm where all alignments subsets
    are searched for every fragment. This is for TIPP, meaning that we also
    perform classification based on a given taxonomy.
    '''
    def __init__(self):
        ExhaustiveAlgorithm.__init__(self)
        self.alignment_threshold = self.options.alignment_threshold
        self.placer = self.options.exhaustive.placer.lower()
        self.push_down = True if self.options.push_down is True else False
        _LOG.info("Will push fragments %s from their placement edge." % (
            "down" if self.push_down else "up"))

    def merge_results(self):
        assert isinstance(self.root_problem, SeppProblem)

        '''Generate single extended alignment'''
        fullExtendedAlignment = ExtendedAlignment(
            self.root_problem.fragments.keys())
        # self.root_problem.get_children()[0].jobs[get_placement_job_name(0)]\
        # .get_attribute("full_extended_alignment_object")
        for pp in self.root_problem.get_children():
            for i in range(0, self.root_problem.fragment_chunks):
                align_input = open(
                    pp.jobs[get_placement_job_name(i)]
                    .full_extended_alignment_file, 'rb')
                extended_alignment = pickle.load(align_input)
                align_input.close()
                fullExtendedAlignment.merge_in(
                    extended_alignment, convert_to_string=True)
        self.results = fullExtendedAlignment

        mergeinput = []
        '''Append main tree to merge input'''
        mergeinput.append("%s;" % (
            self.root_problem.subtree.compose_newick(labels=True)))
        for pp in self.root_problem.get_children():
            assert isinstance(pp, SeppProblem)
            for i in range(0, self.root_problem.fragment_chunks):
                if (pp.get_job_result_by_name(
                       get_placement_job_name(i)) is None):
                    continue
                '''Append subset trees and json locations to merge input'''
                mergeinput.append(
                    "%s;\n%s" % (
                        pp.subtree.compose_newick(labels=True),
                        pp.get_job_result_by_name(get_placement_job_name(i))))
        mergeinput.append("")
        mergeinput.append("")
        meregeinputstring = "\n".join(mergeinput)
        mergeJsonJob = self.get_merge_job(meregeinputstring)
        mergeJsonJob.run()

    def check_options(self, supply=[]):
        if (options().reference_pkg is not None):
            self.load_reference(
                os.path.join(
                    options().reference.path,
                    'refpkg/%s.refpkg/' % options().reference_pkg))
        if (options().taxonomy_file is None):
            supply = supply + ["taxonomy file"]
        if (options().taxonomy_name_mapping_file is None):
            supply = supply + ["taxonomy name mapping file"]
        ExhaustiveAlgorithm.check_options(self, supply)

    def _get_new_Join_Align_Job(self):
        return TIPPJoinAlignJobs(self.placer)

    def build_jobs(self):
        assert isinstance(self.root_problem, SeppProblem)
        for placement_problem in self.root_problem.get_children():
            ''' Create placer jobs'''
            for i in range(0, self.root_problem.fragment_chunks):
                if self.placer == "pplacer":
                    pj = PplacerJob()
                    pj.partial_setup_for_subproblem(
                        placement_problem, self.options.info_file, i)
                elif self.placer == "epa":
                    raise ValueError("EPA Currently not supported")
                    # pj = EPAJob()
                    # pj.partial_setup_for_subproblem(
                    #    placement_problem, self.molecule, i)

                placement_problem.add_job(get_placement_job_name(i), pj)

            '''For each alignment subproblem, ...'''
            for alg_problem in placement_problem.children:
                assert isinstance(alg_problem, SeppProblem)
                ''' create the build model job'''
                bj = HMMBuildJob()
                bj.setup_for_subproblem(alg_problem, molecule=self.molecule)
                alg_problem.add_job(bj.job_type, bj)
                ''' create the search jobs'''
                for fc_problem in alg_problem.get_children():
                    sj = HMMSearchJob()
                    sj.partial_setup_for_subproblem(
                        fc_problem.fragments, fc_problem, self.elim,
                        self.filters)
                    fc_problem.add_job(sj.job_type, sj)
                    ''' create the align job'''
                    aj = HMMAlignJob()
                    fc_problem.add_job(aj.job_type, aj)
                    aj.partial_setup_for_subproblem(
                        fc_problem, molecule=self.molecule)

    def connect_jobs(self):
        ''' a callback function called after hmmbuild jobs are finished'''
        def enq_job_searchfragment(result, search_job):
            search_job.hmmmodel = result
            JobPool().enqueue_job(search_job)
        assert isinstance(self.root_problem, SeppProblem)
        for placement_problem in self.root_problem.get_children():
            '''For each alignment subproblem, ...'''
            for alg_problem in placement_problem.children:
                assert isinstance(alg_problem, SeppProblem)
                ''' create the build model job'''
                bj = alg_problem.jobs["hmmbuild"]
                ''' create the search jobs'''
                for fc_problem in alg_problem.get_children():
                    sj = fc_problem.jobs["hmmsearch"]
                    ''' connect bulid and search jobs'''
                    bj.add_call_Back(
                        lambda result, next_job=sj: enq_job_searchfragment(
                            result, next_job))
            '''Join all align jobs of a placement subset
               (enqueues placement job)'''
            jaj = self._get_new_Join_Align_Job()
            jaj.setup_with_placement_problem(placement_problem)
        ''' Join all search jobs together (enqueues align jobs)'''
        jsj = TIPPJoinSearchJobs(self.alignment_threshold)
        jsj.setup_with_root_problem(self.root_problem)

    def load_reference(self, reference_pkg):
        file = open(reference_pkg + 'CONTENTS.json')
        result = json.load(file)
        file.close()
        options().taxonomy_name_mapping_file = open(
            reference_pkg + result['files']['seq_info'])
        options().taxonomy_file = open(
            reference_pkg + result['files']['taxonomy'])
        options().alignment_file = open(
            reference_pkg + result['files']['aln_fasta'])
        options().tree_file = open(reference_pkg + result['files']['tree'])
        options().info_file = reference_pkg + result['files']['tree_stats']

    def read_alignment_and_tree(self):
        (alignment, tree) = AbstractAlgorithm.read_alignment_and_tree(self)
        # TODO: Check for rooted input
        # if not tree.is_rooted:
        #    raise Exception ("For TIPP, backbone tree should be correctly
        # rooted according to the taxonomy.")
        return (alignment, tree)

    def get_merge_job(self, meregeinputstring):
        mergeJsonJob = TIPPMergeJsonJob()
        mergeJsonJob.setup_for_tipp(
            meregeinputstring,
            self.get_output_filename("placement.json"),
            self.options.taxonomy_file,
            self.options.taxonomy_name_mapping_file,
            self.options.placement_threshold,
            self.get_output_filename("classification.txt"),
            self.push_down,
            self.options.distribution,
            self.options.cutoff)
        return mergeJsonJob

    def get_alignment_decomposition_tree(self, p_tree):
        assert isinstance(p_tree, PhylogeneticTree)
        if self.options.alignment_decomposition_tree is None:
            return PhylogeneticTree(Tree(p_tree.den_tree))
        elif p_tree.count_leaves() != self.root_problem.subtree.count_leaves():
            raise ValueError(
                ("Alignment decomposition tree can be different from placement"
                 " tree only if placement subset size is set to the number of"
                 " taxa (i.e. entire tree)"))
        else:
            _LOG.info(
                "Reading alignment decomposition input tree: %s" % (
                    self.options.alignment_decomposition_tree))
            d_tree = PhylogeneticTree(
                dendropy.Tree.get_from_stream(
                    self.options.alignment_decomposition_tree,
                    schema="newick",
                    preserve_underscores=True,
                    taxon_set=self.root_problem.subtree.get_tree().taxon_set))
            return d_tree


def augment_parser():
    root_p = open(
        os.path.join(
            os.path.split(
                os.path.split(
                    __file__)[0])[0], "home.path")).readlines()[0].strip()
    tipp_config_path = os.path.join(root_p, "tipp.config")
    sepp.config.set_main_config_path(tipp_config_path)
    # default_settings['DEF_P'] = (100 ,
    #    "Number of taxa (i.e. no decomposition)")
    parser = sepp.config.get_parser()
    tippGroup = parser.add_argument_group(
        "TIPP Options".upper(),
        "These arguments set settings specific to TIPP")

    tippGroup.add_argument(
        "-R", "--reference_pkg", type=str,
        dest="reference_pkg", metavar="N",
        default=None,
        help="Use a pre-computed reference package [default: None]")

    tippGroup.add_argument(
        "-at", "--alignmentThreshold", type=float,
        dest="alignment_threshold", metavar="N",
        default=0.95,
        help="Enough alignment subsets are selected to reach a commulative "
             "probability of N. "
             "This should be a number between 0 and 1 [default: 0.95]")
    tippGroup.add_argument(
        "-D", "--dist",
        dest="distribution", action='store_true',
        default=False,
        help="Treat fragments as distribution")

    tippGroup.add_argument(
        "-pt", "--placementThreshold", type=float,
        dest="placement_threshold", metavar="N",
        default=0.95,
        help="Enough placements are selected to reach a commulative "
             "probability of N. "
             "This should be a number between 0 and 1 [default: 0.95]")

    tippGroup.add_argument(
        "-PD", "--push_down", type=bool,
        dest="push_down", metavar="N",
        default=True,
        help="Whether to classify based on children below or above insertion "
             "point.  [default: True]")

    tippGroup.add_argument(
        "-tx", "--taxonomy", type=argparse.FileType('r'),
        dest="taxonomy_file", metavar="TAXONOMY",
        help="A file describing the taxonomy. This is a comma-separated text "
             "file that has the following fields: "
             "taxon_id,parent_id,taxon_name,rank. "
             "If there are other columns, they are ignored."
             " The first line is also ignored.")

    tippGroup.add_argument(
        "-txm", "--taxonomyNameMapping", type=argparse.FileType('r'),
        dest="taxonomy_name_mapping_file", metavar="MAPPING",
        help="A comma-separated text file mapping alignment sequence names to"
             " taxonomic ids. "
             "Formats (each line): "
             "sequence_name,taxon_id. "
             "If there are other columns, they are ignored. The first line is"
             " also ignored.")

    tippGroup.add_argument(
        "-adt", "--alignmentDecompositionTree", type=argparse.FileType('r'),
        dest="alignment_decomposition_tree", metavar="TREE", default=None,
        help="A newick tree file used for decomposing taxa into alignment"
             " subsets. "
             "[default: the backbone tree]")

    tippGroup.add_argument(
        "-C", "--cutoff", type=float,
        dest="cutoff", metavar="N",
        default=0.0,
        help="Placement probability requirement to count toward the"
             " distribution. "
             "This should be a number between 0 and 1 [default: 0.0]")


def main():
    augment_parser()
    TIPPExhaustiveAlgorithm().run()


if __name__ == '__main__':
    main()



'''
Created on Oct 2, 2012

@author: smirarab
'''
import unittest
import sys

from sepp.config import options
from sepp import config
import os
import io
try:
    filetypes = (io.IOBase, file)
except NameError:
    filetypes = io.IOBase
from sepp.filemgr import get_data_path
from tempfile import mkstemp
from sepp.scheduler import JobPool
from multiprocessing import cpu_count


class Test(unittest.TestCase):
    fp_config = None

    def setUp(self):
        _, self.fp_config = mkstemp()

    def tearDown(self):
        os.remove(self.fp_config)

    def testConfigFile(self):
        # Just to make different test cases independent of each other.
        config._options_singelton = None
        # Diasable main config path for this test
        config.main_config_path = self.fp_config

        sys.argv = [
            sys.argv[0], "-A", "2",
            "-c", get_data_path("configs/test.config"),
            "--outdir", "dir_form_commandline"]

        assert options().alignment_size == 2, \
            "Commandline option -A not read properly"

        assert isinstance(options().config_file, filetypes) and \
            options().config_file.name.endswith("data/configs/test.config"), \
            "Commandline option -c not read properly"

        assert (options().pplacer is not None and
                options().pplacer.path == "pplacer"), \
            "config file options not read properly"

        assert options().placement_size == 10, \
            "Config file option placementSize not read properly"

        assert options().outdir.endswith("dir_form_commandline"), \
            "Config file value outdir is not properly overwritten:%s " % \
            options().outdir

        assert options().tempdir is not None, \
            "Default value not properly set for tempfile attribute"

    def testConfigFileMissingFile(self):
        # Just to make different test cases independent of each other.
        config._options_singelton = None
        # Diasable main config path for this test
        config.main_config_path = self.fp_config

        sys.argv = [sys.argv[0],
                    "-c", get_data_path("configs/test2.config"),
                    "-f", get_data_path("simulated/test.fas"),
                    "-a", get_data_path("simulated/test.small.fas")]
        assert isinstance(options().config_file, filetypes) and \
            options().config_file.name.endswith(
                "data/configs/test2.config"), \
            "Commandline option -c not read properly"

        assert isinstance(options().alignment_file, filetypes) and\
            options().alignment_file.name.endswith(
                "data/simulated/test.small.fas"), \
            "Config file option alignment not read properly"

        assert isinstance(options().fragment_file, filetypes) and\
            options().fragment_file.name.endswith(
                "data/simulated/test.fas"), \
            "Command-line option -f alignment not read properly"

    def testMainConfigFile(self):
        # Just to make different test cases independent of each other.
        config._options_singelton = None

        sys.argv = [sys.argv[0], "-c", get_data_path("configs/test2.config")]
        # set pplacer filepath to a file shipped with the code base
        options().pplacer.path = get_data_path(
            "../../../tools/bundled/Darwin/pplacer")

        assert (options().pplacer is not None and os.path.exists(
                options().pplacer.path)), \
            ("main config file options not read properly,"
             "or nonexistent binaries: pplacer = %s" %
             options().pplacer.path)

        options().hmmalign.path = get_data_path(
            "../../../tools/bundled/Darwin/hmmalign")
        assert (options().hmmalign is not None and os.path.exists(
                options().hmmalign.path)), \
            ("main config file options not read properly, or nonexistent "
             "binaries: hmmalign = %s" % options().hmmalign.path)

        options().hmmsearch.path = get_data_path(
            "../../../tools/bundled/Darwin/hmmsearch")
        assert (options().hmmsearch is not None and os.path.exists(
                options().hmmsearch.path)), \
            ("main config file options not read properly, or nonexistent "
             "binaries: hmmsearch = %s" % options().hmmsearch.path)

    def testCpuCount(self):
        # Just to make different test cases independent of each other.
        config._options_singelton = None
        # Disable main config path for this test
        config.main_config_path = self.fp_config
        JobPool().terminate()
        JobPool().__init__(7)
        sys.argv = [sys.argv[0], "-x", "7"]

        assert options().cpu == 7, "Commandline option -x not read properly"

        # clean up after test:
        # 1) the JobPool CPU counts needs to be reset to the default
        # 2) the command line arguments must be restored
        JobPool().terminate()
        JobPool().__init__(cpu_count())
        sys.argv = [sys.argv[0], "-x", str(cpu_count())]
        config._options_singelton = None
        options()

    def testLog(self):
        import sepp
        import logging
        import sepp.jobs

        sdb = sepp._DEBUG

        sepp._DEBUG = True
        sepp.reset_loggers()
        sepp.jobs._LOG.debug("test debugging works")
        assert(sepp.jobs._LOG.getEffectiveLevel() == logging.DEBUG)

        sepp._DEBUG = False
        sepp.reset_loggers()
        sepp.jobs._LOG.debug("test debugging is disabled")
        assert(sepp.jobs._LOG.getEffectiveLevel() == logging.INFO)

        sepp._DEBUG = sdb
        sepp.reset_loggers()


if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testConfigFile']
    unittest.main()
